import datetime
import json
import logging
import math
import os

import requests

from chirp import chirp_common
from chirp import errors
from chirp import platform as chirp_platform
from chirp.sources import base
from chirp.wxui import fips

LOG = logging.getLogger(__name__)

NA_COUNTRIES = [
    'United States',
    'Canada',
    'Mexico',
]
MEXICO_STATES = [
    "Aguascalientes", "Baja California Sur", "Baja California",
    "Campeche", "Chiapas", "Chihuahua", "Coahuila", "Colima",
    "Durango", "Guanajuato", "Guerrero", "Hidalgo", "Jalisco",
    "Mexico City", "Mexico", "Michoacán", "Morelos", "Nayarit",
    "Nuevo Leon", "Puebla", "Queretaro", "Quintana Roo", "San Luis Potosi",
    "Sinaloa", "Sonora", "Tabasco", "Tamaulipas", "Tlaxcala", "Veracruz",
    "Yucatan", "Zacatecas",
]
STATES = {
    'United States': [s for s, i in fips.FIPS_STATES.items()
                      if isinstance(i, int)],
    'Canada': [s for s, i in fips.FIPS_STATES.items()
               if isinstance(i, str)],
    'Mexico': MEXICO_STATES,
}
MODES = ['FM', 'DV', 'DMR', 'DN']


def parse_tone(val):
    if val.startswith('D'):
        mode = 'DTCS'
        val = int(val[1:])
    elif '.' in val:
        mode = 'Tone'
        val = float(val)
    elif val in ('CSQ', 'Restricted'):
        val = mode = None
    elif val:
        LOG.warning('Unsupported PL format: %r' % val)
        val = mode = None
    else:
        val = mode = None
    return mode, val


def distance(lat_a, lon_a, lat_b, lon_b):
    lat_a = math.radians(lat_a)
    lon_a = math.radians(lon_a)

    lat_b = math.radians(lat_b)
    lon_b = math.radians(lon_b)

    earth_radius_km = 6371

    dlon = lon_b - lon_a
    dlat = lat_b - lat_a

    a = math.sin(dlat / 2)**2 + math.cos(lat_a) * \
        math.cos(lat_b) * math.sin(dlon / 2)**2
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    return earth_radius_km * c


class RepeaterBook(base.NetworkResultRadio):
    VENDOR = 'RepeaterBook'

    def get_label(self):
        return 'RepeaterBook'

    def get_data(self, status, country, state, service):
        # Ideally we would be able to pull the whole database, but right
        # now this is limited to 3500 results, so we need to filter and
        # cache by state to stay under that limit.
        fn = 'rb%s-%s-%s.json' % (service,
                                  country.lower().replace(' ', '_'),
                                  state.lower().replace(' ', '_'))
        db_dir = chirp_platform.get_platform().config_file('repeaterbook')
        try:
            os.mkdir(db_dir)
        except FileExistsError:
            pass
        except Exception as e:
            LOG.exception('Failed to create %s: %s' % (db_dir, e))
            status.set_fail('Internal error - check log')
            return
        data_file = os.path.join(db_dir, fn)
        try:
            modified = os.path.getmtime(data_file)
        except FileNotFoundError:
            modified = 0
        modified_dt = datetime.datetime.fromtimestamp(modified)
        interval = datetime.timedelta(days=30)
        if datetime.datetime.now() - modified_dt < interval:
            return data_file
        if modified == 0:
            LOG.debug('RepeaterBook database %s not cached' % fn)
        else:
            LOG.debug('RepeaterBook database %s too old: %s',
                      fn, modified_dt)

        params = {'country': country,
                  'stype': service}
        if country in NA_COUNTRIES:
            export = 'export.php'
        else:
            export = 'exportROW.php'
        if country in STATES:
            params['state'] = state

        r = requests.get('https://www.repeaterbook.com/api/%s' % export,
                         headers=base.HEADERS,
                         params=params,
                         stream=True)
        if r.status_code != 200:
            if modified:
                status.send_status('Using cached data', 50)
            status.send_fail('Got error code %i from server' % r.status_code)
            return
        tmp = data_file + '.tmp'
        chunk_size = 8192
        probable_end = 3 << 20
        counter = 0
        data = b''
        with open(tmp, 'wb') as f:
            for chunk in r.iter_content(chunk_size=chunk_size):
                f.write(chunk)
                data += chunk
                counter += len(chunk)
                status.send_status('Downloading', counter / probable_end * 50)
        try:
            results = json.loads(data)
        except Exception as e:
            LOG.exception('Invalid JSON in response: %s' % e)
            status.send_fail('RepeaterBook returned invalid response')
            return

        if results['count']:
            try:
                os.rename(tmp, data_file)
            except FileExistsError:
                # Windows can't do atomic rename
                os.remove(data_file)
                os.rename(tmp, data_file)
        else:
            os.remove(tmp)
            status.send_fail('No results!')
            return

        status.send_status('Download complete', 50)
        return data_file

    def item_to_memory(self, item, number):
        if item.get('D-Star') == 'Yes':
            m = chirp_common.DVMemory()
            m.dv_urcall = 'CQCQCQ'.ljust(8)
            m.dv_rpt1call = item.get('Callsign')[:8].ljust(8)
            m.dv_rpt2call = item.get('Callsign')[:8].ljust(8)
        else:
            m = chirp_common.Memory()
        m.number = number
        m.freq = chirp_common.parse_freq(item['Frequency'])
        try:
            m.tuning_step = chirp_common.required_step(m.freq)
        except errors.InvalidDataError as e:
            LOG.debug(e)
        txf = chirp_common.parse_freq(item['Input Freq'])
        if txf == 0:
            m.duplex = 'off'
        else:
            chirp_common.split_to_offset(m, m.freq, txf)
        txm, tx = parse_tone(item['PL'])
        rxm, rx = parse_tone(item['TSQ'])
        chirp_common.split_tone_decode(m, (txm, tx, 'N'), (rxm, rx, 'N'))
        if item['DMR'] == 'Yes':
            m.mode = 'DMR'
        elif item['D-Star'] == 'Yes':
            m.mode = 'DV'
        elif item['System Fusion'] == 'Yes':
            m.mode = 'DN'
        elif item['FM Analog'] == 'Yes':
            m.mode = 'FM'
        else:
            LOG.warning('Unable to determine mode for repeater %s' % (
                item['Rptr ID']))
            return None
        if 'State' in item and 'County' in item:
            m.comment = (
                '%(Callsign)s near %(Nearest City)s, %(County)s County, '
                '%(State)s %(Use)s') % item
        else:
            m.comment = (
                '%(Callsign)s near %(Nearest City)s, %(Region)s '
                '%(Use)s') % item
        m.comment += ' ' + item.get('Notes', '')
        m.comment = m.comment.strip()
        m.name = item['Landmark'] or item['Callsign']
        return m

    def do_fetch(self, status, params):
        lat = float(params.pop('lat') or 0)
        lon = float(params.pop('lon') or 0)
        dist = int(params.pop('dist') or 0)
        search_filter = params.pop('filter', '')
        bands = params.pop('bands', [])
        modes = params.pop('modes', [])
        fmconv = params.pop('fmconv', False)
        data_file = self.get_data(status,
                                  params.get('country'),
                                  params.pop('state'),
                                  params.get('service', ''))
        if not data_file:
            return

        status.send_status('Parsing', 50)

        def sorter(item):
            if lat == 0 and lon == 0:
                # No sort if not provided
                return 0
            return distance(lat, lon,
                            float(item.get('Lat', 0)),
                            float(item.get('Long', 0)))

        def match(item):
            search_fields = ('County', 'State', 'Landmark', 'Nearest City',
                             'Callsign', 'Region', 'Notes')
            content = ' '.join(item.get(k) or '' for k in search_fields
                               if k in item)
            return (not search_filter or
                    search_filter.lower() in content.lower())

        def included_band(item):
            if not bands:
                return True
            for lo, hi in bands:
                if lo < chirp_common.parse_freq(item['Frequency']) < hi:
                    return True
            return False

        i = 0
        for item in sorted(json.loads(open(data_file, 'rb').read())['results'],
                           key=sorter):
            if not item:
                continue
            if item['Operational Status'] != 'On-air':
                continue
            if dist and lat and lon and (
                distance(lat, lon,
                         float(item.get('Lat') or 0),
                         float(item.get('Long') or 0)) > dist):
                continue
            if not match(item):
                continue
            if not included_band(item):
                continue
            i += 1
            try:
                m = self.item_to_memory(item, i)
            except Exception as e:
                LOG.warning('Unable to convert repeater %s: %s',
                            item['Rptr ID'], e)
                continue
            if not m:
                continue
            # Convert any non-FM repeater to FM if user requested it
            if m.mode != 'FM' and fmconv and item.get('FM Analog') == 'Yes':
                LOG.debug('Converting repeater %r from %r to FM: %s',
                          item['Rptr ID'], m.mode, m.comment)
                m.mode = 'FM'
            if modes and m.mode not in modes:
                continue
            self._memories.append(m)

        self.MODEL = '%s %s' % (params.get('country'),
                                params.get('service_display') or 'Result')

        if not self._memories:
            status.send_fail(_('No results!'))
            return

        status.send_end()


ROW_COUNTRIES = [
    "Albania",
    "Andorra",
    "Argentina",
    "Australia",
    "Austria",
    "Azerbaijan",
    "Bahamas",
    "Barbados",
    "Belarus",
    "Belgium",
    "Belize",
    "Bolivia",
    "Bosnia and Herzegovina",
    "Brazil",
    "Bulgaria",
    "Caribbean Netherlands",
    "Cayman Islands",
    "Chile",
    "China",
    "Colombia",
    "Costa Rica",
    "Croatia",
    "Curacao",
    "Cyprus",
    "Czech Republic",
    "Denmark",
    "Dominican Republic",
    "Ecuador",
    "El Salvador",
    "Estonia",
    "Faroe Islands",
    "Finland",
    "France",
    "Georgia",
    "Germany",
    "Greece",
    "Grenada",
    "Guatemala",
    "Guernsey",
    "Haiti",
    "Honduras",
    "Hungary",
    "Iceland",
    "India",
    "Indonesia",
    "Ireland",
    "Isle of Man",
    "Israel",
    "Italy",
    "Jamaica",
    "Japan",
    "Jersey",
    "Kosovo",
    "Kuwait",
    "Latvia",
    "Liechtenstein",
    "Lithuania",
    "Luxembourg",
    "Macedonia",
    "Malaysia",
    "Malta",
    "Moldova",
    "Morocco",
    "Namibia",
    "Nepal",
    "Netherlands",
    "New Zealand",
    "Nicaragua",
    "Norway",
    "Oman",
    "Panama",
    "Paraguay",
    "Peru",
    "Philippines",
    "Poland",
    "Portugal",
    "Romania",
    "Russian Federation",
    "Saint Kitts and Nevis",
    "Saint Vincent and the Grenadines",
    "San Marino",
    "Serbia",
    "Singapore",
    "Slovakia",
    "Slovenia",
    "South Africa",
    "South Korea",
    "Spain",
    "Sri Lanka",
    "Sweden",
    "Switzerland",
    "Taiwan",
    "Thailand",
    "Trinidad and Tobago",
    "Turkey",
    "Ukraine",
    "United Arab Emirates",
    "United Kingdom",
    "Uruguay",
    "Venezuela",
]

COUNTRIES = list(sorted(NA_COUNTRIES + ROW_COUNTRIES))



# Copyright 2022 Dan Smith <chirp@f.danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from chirp import chirp_common
from chirp import directory
from chirp.drivers import icf, id31
from chirp import errors
from chirp import bitwise

LOG = logging.getLogger(__name__)


MEM_FORMAT = """
struct {
  u24  mult1:3,
       mult2:3,
       freq:18;
  u16  offset;
  u16  rtone:6,
       ctone:6,
       unknown1:1,
       mode:3;
  u8   dtcs;
  u8   tuning_step:4,
       unknown2:4;
  u8   unknown3;
  u8   tmode:4,
       duplex:2,
       dtcs_polarity:2;
  char name[16];
  u8   dv_code;
  u8   urcall[7];
  u8   rpt1call[7];
  u8   rpt2call[7];
} memory[1004];

#seekto 0xC040;
u8 usedflags[125];
#seekto 0xC0BE;
u8 skipflags[125];
#seekto 0xC13B;
u8 pskipflags[125];

#seekto 0xC1C0;
struct {
  u8 bank;
  u8 index;
} bank_info[1000];

#seekto 0xC9C0;
char ICFComment[16];
struct {
  char name[16];
} bank_names[26];
"""

TUNE_STEPS = [5.0, 6.25, 8.33, 5.0, 10.0, 12.5, 15.0, 20.0, 25.0, 30.0, 50.0,
              5.0, 5.0, 5.0, 5.0]
MODES = ['FM', 'NFM', '2??', 'AM', 'NAM', 'DV', '6??', '7??']
TMODES = ['', 'Tone', '2??', 'TSQL', '4??', 'DTCS', 'TSQL-R', 'DTCS-R',
          'DTCS-T', 'Tone->DTCS', 'DTCS->Tone', 'Tone->Tone']
DUPLEX = ['', '-', '+']
DTCS_POL = ['NN', 'NR', 'RN', 'RR']
SPECIALS = ['144-C0', '144-C1', '430-C0', '430-C1']
MULTS = [5000, 6250, 25000 / 3.0]


@directory.register
class ID4100Radio(icf.IcomCloneModeRadio, chirp_common.IcomDstarSupport):
    VENDOR = 'Icom'
    MODEL = 'ID-4100'
    NEEDS_COMPAT_SERIAL = False

    _model = b'\x38\x66\x00\x01'
    _endframe = 'Icom Inc.8F'
    _memsize = 0x2A3C0
    _ranges = [(0, _memsize, 64)]
    _raw_frames = True
    _highbit_flip = True

    _num_banks = 26
    _bank_class = id31.ID31Bank

    _can_hispeed = True
    _icf_data = {
        'MapRev': 1,
        'EtcData': 0x400001,
    }

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_ctone = True
        rf.has_dtcs = True
        rf.has_dtcs_polarity = True
        rf.has_bank = True
        rf.has_bank_index = True
        rf.has_bank_names = True
        rf.requires_call_lists = False
        rf.valid_modes = [x for x in MODES
                          if '?' not in x]
        rf.valid_tmodes = [x for x in TMODES
                           if '-' not in x and '?' not in x] + ['Cross']
        rf.valid_cross_modes = [x for x in TMODES
                                if '->' in x]
        rf.memory_bounds = (0, 999)
        rf.valid_bands = [(118000000, 174000000),
                          (375000000, 550000000)]
        rf.valid_skips = ['', 'S', 'P']
        rf.valid_characters = chirp_common.CHARSET_ASCII
        rf.valid_name_length = 16
        rf.valid_tuning_steps = list(sorted(set(TUNE_STEPS)))
        rf.valid_special_chans = list(SPECIALS)
        return rf

    def _get_bank(self, loc):
        _bank = self._memobj.bank_info[loc]
        _bank.bank &= 0x1F
        if _bank.bank == 0x1F:
            return None
        else:
            return _bank.bank

    def _set_bank(self, loc, bank):
        _bank = self._memobj.bank_info[loc]
        if bank is None:
            _bank.bank = 0x1F
        else:
            _bank.bank = bank

    def _get_bank_index(self, loc):
        _bank = self._memobj.bank_info[loc]
        return _bank.index

    def _set_bank_index(self, loc, index):
        _bank = self._memobj.bank_info[loc]
        _bank.index = index

    def _get_raw_memory(self, number):
        if isinstance(number, str):
            number = 1000 + SPECIALS.index(number)
        return number, self._memobj.memory[number]

    def get_raw_memory(self, number):
        num, mem = self._get_raw_memory(number)
        return repr(mem)

    def get_memory(self, number):
        num, _mem = self._get_raw_memory(number)
        m = chirp_common.DVMemory()
        m.number = num
        if isinstance(number, str):
            m.extd_number = number
        else:
            _flg = self._memobj.usedflags[num // 8]
            _skp = self._memobj.skipflags[num // 8]
            _pskp = self._memobj.pskipflags[num // 8]
            m.empty = bool(_flg & (1 << num % 8))
            if _pskp & 1 << (num % 8):
                m.skip = 'P'
            elif _skp & 1 << (num % 8):
                m.skip = 'S'
            else:
                m.skip = ''

        mult = MULTS[_mem.mult1]
        m.freq = int(_mem.freq * mult)
        m.offset = int(_mem.offset * mult)
        m.tuning_step = TUNE_STEPS[_mem.tuning_step]
        m.name = str(_mem.name).rstrip()
        m.mode = MODES[_mem.mode]
        m.rtone = chirp_common.TONES[_mem.rtone]
        m.ctone = chirp_common.TONES[_mem.ctone]
        m.dtcs = chirp_common.DTCS_CODES[_mem.dtcs]
        m.dtcs_polarity = DTCS_POL[_mem.dtcs_polarity]
        tmode = TMODES[_mem.tmode]
        if '->' in tmode:
            m.tmode = 'Cross'
            m.cross_mode = tmode
        elif '-' in tmode and 0:
            # FIXME
            m.tmode, extra = tmode.split('-')
        else:
            m.tmode = tmode
        m.duplex = DUPLEX[_mem.duplex]

        m.dv_code = _mem.dv_code
        m.dv_urcall = ''.join(
            chr(x) for x in icf.warp_byte_size(_mem.urcall, 7, 8))
        m.dv_rpt1call = ''.join(
            chr(x) for x in icf.warp_byte_size(_mem.rpt1call, 7, 8))
        m.dv_rpt2call = ''.join(
            chr(x) for x in icf.warp_byte_size(_mem.rpt2call, 7, 8))

        return m

    def set_memory(self, mem):
        num, _mem = self._get_raw_memory(mem.number)
        if num < 1000:
            _flg = self._memobj.usedflags[num // 8]
            _skp = self._memobj.skipflags[num // 8]
            _pskp = self._memobj.pskipflags[num // 8]
            mybit = 1 << (num % 8)
            _flg |= mybit
            _skp &= ~mybit
            _pskp &= ~mybit

            if mem.empty:
                return

            _flg &= ~mybit
            if mem.skip == 'S':
                _skp |= mybit
            elif mem.skip == 'P':
                _pskp |= mybit

        if chirp_common.is_6_25(mem.freq):
            mult = MULTS[1]
        elif chirp_common.is_8_33(mem.freq):
            mult = MULTS[2]
        else:
            mult = MULTS[0]
        _mem.mult1 = _mem.mult2 = MULTS.index(mult)
        _mem.freq = mem.freq / mult
        _mem.offset = mem.offset / mult
        _mem.tuning_step = TUNE_STEPS.index(mem.tuning_step)
        _mem.name = mem.name.ljust(16)
        _mem.mode = MODES.index(mem.mode)
        _mem.rtone = chirp_common.TONES.index(mem.rtone)
        _mem.ctone = chirp_common.TONES.index(mem.ctone)
        _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)
        _mem.dtcs_polarity = DTCS_POL.index(mem.dtcs_polarity)
        if mem.tmode == 'Cross':
            _mem.tmode = TMODES.index(mem.cross_mode)
        else:
            _mem.tmode = TMODES.index(mem.tmode)
        _mem.duplex = DUPLEX.index(mem.duplex)

        _mem.unknown1 = 0
        _mem.unknown2 = 0
        _mem.unknown3 = 0

        if isinstance(mem, chirp_common.DVMemory):
            _mem.dv_code = mem.dv_code
            _mem.urcall = list(
                icf.warp_byte_size(mem.dv_urcall.ljust(8), 8, 7))
            _mem.rpt1call = list(
                icf.warp_byte_size(mem.dv_rpt1call.ljust(8), 8, 7))
            _mem.rpt2call = list(
                icf.warp_byte_size(mem.dv_rpt2call.ljust(8), 8, 7))


@directory.register
class ID5100Radio(ID4100Radio):
    MODEL = "ID-5100"

    _model = b'\x34\x84\x00\x01'

    # This is only for MapRev=1
    _endframe = "Icom Inc.EE"

    # MapRev=1 Size is 260928 0x3FB40
    # MapRev=2 Size is 148160 0x242C0
    # MapRev=3 Size is 172928 0x2A380
    _memsize = 0x3FB40

    _ranges = [(0, _memsize, 64)]
    _raw_frames = True
    _highbit_flip = True

    def process_mmap(self):
        was_memsize = self._memsize

        # Apparently the 5100 has different reported memory sizes
        # depending on firmware version. When we're loading an image,
        # we should adjust our MapRev and _ranges to be correct for
        # saving ICF files and uploading. This also means we should
        # try to detect which version we are talking to and refuse to
        # upload an image to a radio with a mismatch.

        self._memsize = len(self._mmap)
        self._ranges = [(0, self._memsize, 64)]

        # Major (memory-format-changing) firmware versions, which
        # correspond to MapRev in an ICF file.
        maprevs = {
            0x3FB40: 1,
            0x242C0: 2,
            0x2A380: 3,
        }

        # The 5100 seems to almost behave like three different radios,
        # depending on firmware version. These are the endframes that
        # are expected for a given MapRev.
        endframes = {
            1: 'Icom Inc.EE',
            2: 'Icom Inc.0C',
            3: 'Icom Inc.8E',
        }

        if self._memsize != was_memsize:
            self._icf_data['MapRev'] = maprevs.get(self._memsize, 0)
            if self._icf_data['MapRev'] == 0:
                LOG.error('Unknown memsize %06X!', self._memsize)
                raise errors.InvalidDataError('Unsupported memory format!')
            self._endframe = endframes[self._icf_data['MapRev']]
            LOG.info('Memory length changed from %06X to %06X; new MapRev=%i',
                     was_memsize, self._memsize, self._icf_data['MapRev'])
        else:
            LOG.debug('Unchanged memsize at %06X' % self._memsize)

        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.info = _('This driver has been tested with v3 of the ID-5100. '
                    'If your radio is not fully updated please help by '
                    'opening a bug report with a debug log so we can add '
                    'support for the other revisions.')

        rp.pre_upload = rp.info
        rp.experimental = rp.info
        return rp



# Copyright 2011 Dan Smith <dsmith@danplanet.com>
# Copyright 2013 Jens Jensen AF5MI <kd4tjx@yahoo.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp.drivers import yaesu_clone
from chirp import chirp_common, bitwise, memmap, directory, errors, util
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueList, RadioSettingValueBoolean, \
    RadioSettingValueString, RadioSettings

import time
import traceback
import string
import re
import logging


LOG = logging.getLogger(__name__)

CMD_ACK = b'\x06'

FT90_STEPS = [5.0, 10.0, 12.5, 15.0, 20.0, 25.0, 50.0]
FT90_MODES = ["AM", "FM", "Auto"]
# idx 3 (Bell) not supported yet
FT90_TMODES = ["", "Tone", "TSQL", "", "DTCS"]
FT90_TONES = list(chirp_common.TONES)
for tone in [165.5, 171.3, 177.3]:
    FT90_TONES.remove(tone)
FT90_POWER_LEVELS_VHF = [chirp_common.PowerLevel("Hi", watts=50),
                         chirp_common.PowerLevel("Mid1", watts=20),
                         chirp_common.PowerLevel("Mid2", watts=10),
                         chirp_common.PowerLevel("Low", watts=5)]

FT90_POWER_LEVELS_UHF = [chirp_common.PowerLevel("Hi", watts=35),
                         chirp_common.PowerLevel("Mid1", watts=20),
                         chirp_common.PowerLevel("Mid2", watts=10),
                         chirp_common.PowerLevel("Low", watts=5)]

FT90_DUPLEX = ["", "-", "+", "split"]
FT90_CWID_CHARS = list(string.digits) + list(string.ascii_uppercase) \
                  + list(" ")
FT90_DTMF_CHARS = list("0123456789ABCD*#")
FT90_SPECIAL = ["vfo_vhf", "home_vhf", "vfo_uhf", "home_uhf",
                "pms_1L", "pms_1U", "pms_2L", "pms_2U"]


@directory.register
class FT90Radio(yaesu_clone.YaesuCloneModeRadio):
    VENDOR = "Yaesu"
    MODEL = "FT-90"
    ID = b"\x8E\xF6"
    NEEDS_COMPAT_SERIAL = False

    _memsize = 4063
    # block 03 (200 Bytes long) repeats 18 times; channel memories
    _block_lengths = [2, 232, 24] + ([200] * 18) + [205]

    mem_format = """
u16 id;
#seekto 0x22;
struct {
    u8 dtmf_active;
    u8 dtmf1_len;
    u8 dtmf2_len;
    u8 dtmf3_len;
    u8 dtmf4_len;
    u8 dtmf5_len;
    u8 dtmf6_len;
    u8 dtmf7_len;
    u8 dtmf8_len;
    u8 dtmf1[8];
    u8 dtmf2[8];
    u8 dtmf3[8];
    u8 dtmf4[8];
    u8 dtmf5[8];
    u8 dtmf6[8];
    u8 dtmf7[8];
    u8 dtmf8[8];
    char cwid[7];
    u8 unk1;
    u8 scan1:2,
       beep:1,
       unk3:3,
       rfsqlvl:2;
    u8 unk4:2,
       scan2:1,
       cwid_en:1,
       txnarrow:1,
       dtmfspeed:1,
       pttlock:2;
    u8 dtmftxdelay:3,
       fancontrol:2,
       unk5:3;
    u8 dimmer:3,
       unk6:1,
       lcdcontrast:4;
    u8 dcsmode:2,
       unk16:2,
       tot:4;
    u8 unk14;
    u8 unk8:1,
       ars:1,
       lock:1,
       txpwrsave:1,
       apo:4;
    u8 unk15;
    u8 unk9:4,
       key_lt:4;
    u8 unk10:4,
       key_rt:4;
    u8 unk11:4,
       key_p1:4;
    u8 unk12:4,
       key_p2:4;
    u8 unk13:4,
       key_acc:4;
} settings;

struct mem_struct {
    u8 mode:2,
       isUhf1:1,
       unknown1:2,
       step:3;
    u8 artsmode:2,
       unknown2:1,
       isUhf2:1,
       power:2,
       shift:2;
    u8 skip:1,
       showname:1,
       unknown3:1,
       isUhfHi:1,
       unknown4:1,
       tmode:3;
    u32 rxfreq;
    u32 txfreqoffset;
    u8 UseDefaultName:1,
       ars:1,
       tone:6;
    u8 packetmode:1,
       dcstone:7;
    char name[7];
};

#seekto 0x86;
struct mem_struct vfo_vhf;
struct mem_struct home_vhf;
struct mem_struct vfo_uhf;
struct mem_struct home_uhf;

#seekto 0xEB;
u8 chan_enable[23];

#seekto 0x101;
struct {
    u8 pms_2U_enable:1,
       pms_2L_enable:1,
       pms_1U_enable:1,
       pms_1L_enable:1,
       unknown6:4;
} special_enables;

#seekto 0x102;
struct mem_struct memory[180];

#seekto 0xf12;
struct mem_struct pms_1L;
struct mem_struct pms_1U;
struct mem_struct pms_2L;
struct mem_struct pms_2U;

#seekto 0x0F7B;
struct  {
    char demomsg1[50];
    char demomsg2[50];
} demomsg;
"""

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.pre_download = _(
            "1. Turn radio off.\n"
            "2. Connect mic and hold [ACC] on mic while powering on.\n"
            "    (\"CLONE\" will appear on the display)\n"
            "3. Replace mic with PC programming cable.\n"
            "4. <b>After clicking OK</b>, press the [SET] key to send"
            " image.\n")
        rp.pre_upload = _(
            "1. Turn radio off.\n"
            "2. Connect mic and hold [ACC] on mic while powering on.\n"
            "    (\"CLONE\" will appear on the display)\n"
            "3. Replace mic with PC programming cable.\n"
            "4. Press the [DISP/SS] key\n"
            "    (\"R\" will appear on the lower left of LCD).\n")
        rp.display_pre_upload_prompt_before_opening_port = False
        return rp

    @classmethod
    def match_model(cls, filedata, filename):
        return len(filedata) == cls._memsize

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_ctone = False
        rf.has_bank = False
        rf.has_dtcs_polarity = False
        rf.has_dtcs = True
        rf.can_odd_split = True
        rf.valid_modes = FT90_MODES
        rf.valid_tmodes = FT90_TMODES
        rf.valid_duplexes = FT90_DUPLEX
        rf.valid_tuning_steps = FT90_STEPS[1:]
        rf.valid_power_levels = FT90_POWER_LEVELS_VHF
        rf.valid_name_length = 7
        rf.valid_characters = chirp_common.CHARSET_ASCII
        rf.valid_skips = ["", "S"]
        rf.valid_special_chans = FT90_SPECIAL
        rf.valid_tones = FT90_TONES
        rf.memory_bounds = (1, 180)
        rf.valid_bands = [(100000000, 230000000),
                          (300000000, 530000000),
                          (810000000, 999975000)]

        return rf

    def _read(self, blocksize, blocknum):
        data = self.pipe.read(blocksize+2)

        # chew echo'd ack
        self.pipe.write(CMD_ACK)
        time.sleep(0.02)
        self.pipe.read(1)  # chew echoed ACK from 1-wire serial

        if len(data) == blocksize + 2 and data[0] == blocknum:
            checksum = yaesu_clone.YaesuChecksum(1, blocksize)
            if checksum.get_existing(data) != checksum.get_calculated(data):
                raise Exception("Checksum Failed [%02X<>%02X] block %02X, "
                                "data len: %i" %
                                (checksum.get_existing(data),
                                 checksum.get_calculated(data),
                                 blocknum, len(data)))
            data = data[1:blocksize + 1]  # Chew blocknum and checksum

        else:
            raise Exception("Unable to read blocknum %02X "
                            "expected blocksize %i got %i." %
                            (blocknum, blocksize+2, len(data)))

        return data

    def _clone_in(self):
        # Be very patient with the radio
        self.pipe.timeout = 4
        start = time.time()

        data = b""
        blocknum = 0
        status = chirp_common.Status()
        status.msg = "Cloning..."
        self.status_fn(status)
        status.max = len(self._block_lengths)
        for blocksize in self._block_lengths:
            data += self._read(blocksize, blocknum)
            blocknum += 1
            status.cur = blocknum
            self.status_fn(status)

        LOG.info("Clone completed in %i seconds, blocks read: %i" %
                 (time.time() - start, blocknum))

        return memmap.MemoryMapBytes(data)

    def _clone_out(self):
        looppredelay = 0.4
        looppostdelay = 1.9
        self.pipe.timeout = 4
        start = time.time()

        blocknum = 0
        pos = 0
        status = chirp_common.Status()
        status.msg = "Cloning to radio..."
        self.status_fn(status)
        status.max = len(self._block_lengths)

        for blocksize in self._block_lengths:
            checksum = yaesu_clone.YaesuChecksum(pos, pos+blocksize-1)
            blocknumbyte = bytes([blocknum])
            payloadbytes = self.get_mmap()[pos:pos+blocksize]
            checksumbyte = bytes([checksum.get_calculated(self.get_mmap())])
            LOG.debug("Block %i - will send from %i to %i byte " %
                      (blocknum, pos, pos + blocksize))
            LOG.debug(util.hexprint(blocknumbyte))
            LOG.debug(util.hexprint(payloadbytes))
            LOG.debug(util.hexprint(checksumbyte))
            # send wrapped bytes
            time.sleep(looppredelay)
            self.pipe.write(blocknumbyte)
            self.pipe.write(payloadbytes)
            self.pipe.write(checksumbyte)
            tmp = self.pipe.read(blocksize + 2)  # chew echo
            LOG.debug("bytes echoed: ")
            LOG.debug(util.hexprint(tmp))
            # radio is slow to write/ack:
            time.sleep(looppostdelay)
            buf = self.pipe.read(1)
            LOG.debug("ack recd:")
            LOG.debug(util.hexprint(buf))
            if buf != CMD_ACK:
                raise Exception("Radio did not ack block %i" % blocknum)
            pos += blocksize
            blocknum += 1
            status.cur = blocknum
            self.status_fn(status)

        LOG.info("Clone completed in %i seconds" % (time.time() - start))

    def sync_in(self):
        try:
            self._mmap = self._clone_in()
        except errors.RadioError:
            raise
        except Exception:
            trace = traceback.format_exc()
            raise errors.RadioError(
                    "Failed to communicate with radio: %s" % trace)
        self.process_mmap()

    def sync_out(self):
        try:
            self._clone_out()
        except errors.RadioError:
            raise
        except Exception:
            trace = traceback.format_exc()
            raise errors.RadioError(
                    "Failed to communicate with radio: %s" % trace)

    def process_mmap(self):
        self._memobj = bitwise.parse(self.mem_format, self._mmap)

    def _get_chan_enable(self, number):
        number = number - 1
        bytepos = number // 8
        bitpos = number % 8
        chan_enable = self._memobj.chan_enable[bytepos]
        if chan_enable & (1 << bitpos):
            return True
        else:
            return False

    def _set_chan_enable(self, number, enable):
        number = number - 1
        bytepos = number // 8
        bitpos = number % 8
        chan_enable = self._memobj.chan_enable[bytepos]
        if enable:
            chan_enable = chan_enable | (1 << bitpos)  # enable
        else:
            chan_enable = chan_enable & ~(1 << bitpos)  # disable
        self._memobj.chan_enable[bytepos] = chan_enable

    def get_memory(self, number):
        mem = chirp_common.Memory()
        if isinstance(number, str):
            # special channel
            _mem = getattr(self._memobj, number)
            mem.number = - len(FT90_SPECIAL) + FT90_SPECIAL.index(number)
            mem.extd_number = number
            if re.match('^pms', mem.extd_number):
                # enable pms_XY channel flag
                _special_enables = self._memobj.special_enables
                # Disabled in 2023 because this doesn't pass tests and I'm
                # not sure what it is really trying to accomplish.
                # mem.empty = not getattr(_special_enables,
                #                         mem.extd_number + "_enable")
        else:
            # regular memory
            _mem = self._memobj.memory[number-1]
            mem.number = number
            mem.empty = not self._get_chan_enable(number)
        if mem.empty:
            return mem  # bail out, do not parse junk
        mem.freq = _mem.rxfreq * 10
        mem.offset = _mem.txfreqoffset * 10
        if not _mem.tmode < len(FT90_TMODES):
            _mem.tmode = 0
        mem.tmode = FT90_TMODES[_mem.tmode]
        mem.rtone = FT90_TONES[_mem.tone]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dcstone]
        mem.mode = FT90_MODES[_mem.mode]
        mem.duplex = FT90_DUPLEX[_mem.shift]
        if mem.freq / 1000000 > 300:
            mem.power = FT90_POWER_LEVELS_UHF[_mem.power]
        else:
            mem.power = FT90_POWER_LEVELS_VHF[_mem.power]

        # radio has a known bug with 5 kHz step and squelch
        if _mem.step == 0 or _mem.step > len(FT90_STEPS)-1:
            _mem.step = 2
        mem.tuning_step = FT90_STEPS[_mem.step]
        mem.skip = _mem.skip and "S" or ""
        if not all(char in chirp_common.CHARSET_ASCII
                   for char in str(_mem.name)):
            # don't display blank/junk name
            mem.name = ""
        else:
            mem.name = str(_mem.name).rstrip()
        return mem

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number-1])

    def set_memory(self, mem):
        if mem.number < 0:      # special channels
            _mem = getattr(self._memobj, mem.extd_number)
            if re.match('^pms', mem.extd_number):
                # enable pms_XY channel flag
                _special_enables = self._memobj.special_enables
                setattr(_special_enables, mem.extd_number + "_enable", True)
        else:
            _mem = self._memobj.memory[mem.number - 1]
            self._set_chan_enable(mem.number, not mem.empty)
        _mem.skip = mem.skip == "S"
        # radio has a known bug with 5 kHz step and dead squelch
        if not mem.tuning_step or mem.tuning_step == FT90_STEPS[0]:
            _mem.step = 2
        else:
            _mem.step = FT90_STEPS.index(mem.tuning_step)
        _mem.rxfreq = mem.freq / 10
        # vfo will unlock if not in right band?
        if mem.freq > 300000000:
            # uhf
            _mem.isUhf1 = 1
            _mem.isUhf2 = 1
            if mem.freq > 810000000:
                # uhf hiband
                _mem.isUhfHi = 1
            else:
                _mem.isUhfHi = 0
        else:
            # vhf
            _mem.isUhf1 = 0
            _mem.isUhf2 = 0
            _mem.isUhfHi = 0
        _mem.txfreqoffset = mem.offset / 10
        _mem.tone = FT90_TONES.index(mem.rtone)
        _mem.tmode = FT90_TMODES.index(mem.tmode)
        _mem.mode = FT90_MODES.index(mem.mode)
        _mem.shift = FT90_DUPLEX.index(mem.duplex)
        _mem.dcstone = chirp_common.DTCS_CODES.index(mem.dtcs)
        _mem.step = FT90_STEPS.index(mem.tuning_step)
        _mem.shift = FT90_DUPLEX.index(mem.duplex)
        if mem.power:
            try:
                _mem.power = FT90_POWER_LEVELS_VHF.index(mem.power)
            except ValueError:
                _mem.power = FT90_POWER_LEVELS_UHF.index(mem.power)
        else:
            _mem.power = 3  # default to low power
        if (len(mem.name) == 0):
            _mem.name = bytearray.fromhex("80ffffffffffff")
            _mem.showname = 0
        else:
            _mem.name = str(mem.name).ljust(7)
            _mem.showname = 1
            _mem.UseDefaultName = 0

    def _decode_cwid(self, cwidarr):
        cwid = ""
        LOG.debug("@ +_decode_cwid:")
        for byte in cwidarr.get_value():
            char = int(byte)
            LOG.debug(char)
            # bitwise wraps in quotes! get rid of those
            if char < len(FT90_CWID_CHARS):
                cwid += FT90_CWID_CHARS[char]
        return cwid

    def _encode_cwid(self, cwidarr):
        cwid = ""
        LOG.debug("@ _encode_cwid:")
        for char in cwidarr.get_value():
            cwid += chr(FT90_CWID_CHARS.index(char))
        LOG.debug(cwid)
        return cwid

    def _bbcd2dtmf(self, bcdarr, strlen=16):
        # doing bbcd, but with support for ABCD*#
        LOG.debug(bcdarr.get_value())
        string = ''.join("%02X" % b for b in bcdarr)
        LOG.debug("@_bbcd2dtmf, received: %s" % string)
        string = string.replace('E', '*').replace('F', '#')
        if strlen <= 16:
            string = string[:strlen]
        return string

    def _dtmf2bbcd(self, dtmf):
        dtmfstr = dtmf.get_value()
        dtmfstr = dtmfstr.replace('*', 'E').replace('#', 'F')
        dtmfstr = str.ljust(dtmfstr.strip(), 16, "0")
        bcdarr = list(bytearray.fromhex(dtmfstr))
        LOG.debug("@_dtmf2bbcd, sending: %s" % bcdarr)
        return bcdarr

    def get_settings(self):
        _settings = self._memobj.settings
        basic = RadioSettingGroup("basic", "Basic")
        autodial = RadioSettingGroup("autodial", "AutoDial")
        keymaps = RadioSettingGroup("keymaps", "KeyMaps")

        top = RadioSettings(basic, keymaps, autodial)

        rs = RadioSetting(
                "beep", "Beep",
                RadioSettingValueBoolean(_settings.beep))
        basic.append(rs)
        rs = RadioSetting(
                "lock", "Lock",
                RadioSettingValueBoolean(_settings.lock))
        basic.append(rs)
        rs = RadioSetting(
                "ars", "Auto Repeater Shift",
                RadioSettingValueBoolean(_settings.ars))
        basic.append(rs)
        rs = RadioSetting(
                "txpwrsave", "TX Power Save",
                RadioSettingValueBoolean(_settings.txpwrsave))
        basic.append(rs)
        rs = RadioSetting(
                "txnarrow", "TX Narrow",
                RadioSettingValueBoolean(_settings.txnarrow))
        basic.append(rs)
        options = ["Off", "S-3", "S-5", "S-Full"]
        rs = RadioSetting(
                "rfsqlvl", "RF Squelch Level",
                RadioSettingValueList(options, options[_settings.rfsqlvl]))
        basic.append(rs)
        options = ["Off", "Band A", "Band B", "Both"]
        rs = RadioSetting(
                "pttlock", "PTT Lock",
                RadioSettingValueList(options, options[_settings.pttlock]))
        basic.append(rs)

        rs = RadioSetting(
                "cwid_en", "CWID Enable",
                RadioSettingValueBoolean(_settings.cwid_en))
        basic.append(rs)

        cwid = RadioSettingValueString(0, 7, self._decode_cwid(_settings.cwid))
        cwid.set_charset(FT90_CWID_CHARS)
        rs = RadioSetting("cwid", "CWID", cwid)
        basic.append(rs)

        options = ["OFF"] + [str(x) for x in (range(1, 12+1))]
        rs = RadioSetting(
                "apo", "APO time (hrs)",
                RadioSettingValueList(options, options[_settings.apo]))
        basic.append(rs)

        options = ["Off"] + [str(x) for x in (range(1, 60+1))]
        rs = RadioSetting(
                "tot", "Time Out Timer (mins)",
                RadioSettingValueList(options, options[_settings.tot]))
        basic.append(rs)

        options = ["off", "Auto/TX", "Auto", "TX"]
        rs = RadioSetting(
                "fancontrol", "Fan Control",
                RadioSettingValueList(options, options[_settings.fancontrol]))
        basic.append(rs)

        keyopts = ["Scan Up", "Scan Down", "Repeater", "Reverse", "Tone Burst",
                   "Tx Power", "Home Ch", "VFO/MR", "Tone", "Priority"]
        rs = RadioSetting(
                "key_lt", "Left Key",
                RadioSettingValueList(keyopts, keyopts[_settings.key_lt]))
        keymaps.append(rs)
        rs = RadioSetting(
                "key_rt", "Right Key",
                RadioSettingValueList(keyopts, keyopts[_settings.key_rt]))
        keymaps.append(rs)
        rs = RadioSetting(
                "key_p1", "P1 Key",
                RadioSettingValueList(keyopts, keyopts[_settings.key_p1]))
        keymaps.append(rs)
        rs = RadioSetting(
                "key_p2", "P2 Key",
                RadioSettingValueList(keyopts, keyopts[_settings.key_p2]))
        keymaps.append(rs)
        rs = RadioSetting(
                "key_acc", "ACC Key",
                RadioSettingValueList(keyopts, keyopts[_settings.key_acc]))
        keymaps.append(rs)

        options = [str(x) for x in range(0, 12+1)]
        rs = RadioSetting(
                "lcdcontrast", "LCD Contrast",
                RadioSettingValueList(options, options[_settings.lcdcontrast]))
        basic.append(rs)

        options = ["off", "d4", "d3", "d2", "d1"]
        rs = RadioSetting(
                "dimmer", "Dimmer",
                RadioSettingValueList(options, options[_settings.dimmer]))
        basic.append(rs)

        options = ["TRX Normal", "RX Reverse", "TX Reverse", "TRX Reverse"]
        rs = RadioSetting(
                "dcsmode", "DCS Mode",
                RadioSettingValueList(options, options[_settings.dcsmode]))
        basic.append(rs)

        options = ["50 ms", "100 ms"]
        rs = RadioSetting(
                "dtmfspeed", "DTMF Speed",
                RadioSettingValueList(options, options[_settings.dtmfspeed]))
        autodial.append(rs)

        options = ["50 ms", "250 ms", "450 ms", "750 ms", "1 sec"]
        rs = RadioSetting(
                "dtmftxdelay", "DTMF TX Delay",
                RadioSettingValueList(options, options[_settings.dtmftxdelay]))
        autodial.append(rs)

        options = [str(x) for x in range(1, 8 + 1)]
        rs = RadioSetting(
                "dtmf_active", "DTMF Active",
                RadioSettingValueList(options, options[_settings.dtmf_active]))
        autodial.append(rs)

        # setup 8 dtmf autodial entries
        for i in range(1, 9):
            objname = "dtmf" + str(i)
            dtmfsetting = getattr(_settings, objname)
            dtmflen = getattr(_settings, objname + "_len")
            dtmfstr = self._bbcd2dtmf(dtmfsetting, dtmflen)
            dtmf = RadioSettingValueString(0, 16, dtmfstr)
            dtmf.set_charset(FT90_DTMF_CHARS + list(" "))
            rs = RadioSetting(objname, objname.upper(), dtmf)
            autodial.append(rs)

        return top

    def set_settings(self, uisettings):
        _settings = self._memobj.settings
        for element in uisettings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue
            try:
                setting = element.get_name()
                oldval = getattr(_settings, setting)
                newval = element.value
                if setting == "cwid":
                    newval = self._encode_cwid(newval)
                if re.match(r'dtmf\d', setting):
                    # set dtmf length field and then get bcd dtmf
                    dtmfstrlen = len(str(newval).strip())
                    setattr(_settings, setting + "_len", dtmfstrlen)
                    newval = self._dtmf2bbcd(newval)
                LOG.debug("Setting %s(%s) <= %s" % (setting, oldval, newval))
                setattr(_settings, setting, newval)
            except Exception:
                LOG.debug(element.get_name())
                raise



# Copyright 2008 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from pathlib import Path
import sys
import re
import logging
from subprocess import Popen

import six

LOG = logging.getLogger(__name__)


def _find_me():
    return sys.modules["chirp.platform"].__file__


def natural_sorted(lst):
    def convert(text):
        return int(text) if text.isdigit() else text.lower()

    def natural_key(key):
        return [convert(c) for c in re.split('([0-9]+)', key)]

    return sorted(lst, key=natural_key)


class Platform:
    """Base class for platform-specific functions"""

    def __init__(self, basepath):
        self._base = basepath
        self._last_dir = self.default_dir()

    def get_last_dir(self):
        """Return the last directory used"""
        return self._last_dir

    def set_last_dir(self, last_dir):
        """Set the last directory used"""
        self._last_dir = last_dir

    def config_dir(self):
        """Return the preferred configuration file directory"""
        return self._base

    def log_dir(self):
        """Return the preferred log file directory"""
        logdir = os.path.join(self.config_dir(), "logs")
        try:
            os.mkdir(logdir)
        except FileExistsError:
            pass

        return logdir

    def filter_filename(self, filename):
        """Filter @filename for platform-forbidden characters"""
        return filename

    def log_file(self, filename):
        """Return the full path to a log file with @filename"""
        filename = self.filter_filename(filename + ".txt").replace(" ", "_")
        return os.path.join(self.log_dir(), filename)

    def config_file(self, filename):
        """Return the full path to a config file with @filename"""
        return os.path.join(self.config_dir(),
                            self.filter_filename(filename))

    def open_text_file(self, path):
        """Spawn the necessary program to open a text file at @path"""
        raise NotImplementedError("The base class can't do that")

    def open_html_file(self, path):
        """Spawn the necessary program to open an HTML file at @path"""
        raise NotImplementedError("The base class can't do that")

    def default_dir(self):
        """Return the default directory for this platform"""
        return "."

    def os_version_string(self):
        """Return a string that describes the OS/platform version"""
        return "Unknown Operating System"

    def executable_path(self):
        """Return a full path to the program executable"""
        def we_are_frozen():
            return hasattr(sys, "frozen")

        if we_are_frozen():
            # Win32, find the directory of the executable
            return os.path.dirname(six.text_type(sys.executable,
                                                 sys.getfilesystemencoding()))
        else:
            # UNIX: Find the parent directory of this module
            return os.path.dirname(os.path.abspath(os.path.join(_find_me(),
                                                                "..")))

    def find_resource(self, filename):
        """Searches for files installed to a share/ prefix."""
        execpath = self.executable_path()
        share_candidates = [
            os.path.join(execpath, "share"),
            os.path.join(sys.prefix, "share"),
            "/usr/local/share",
            "/usr/share",
        ]
        pkgshare_candidates = [os.path.join(i, "chirp")
                               for i in share_candidates]
        search_paths = [execpath] + pkgshare_candidates + share_candidates
        for path in search_paths:
            candidate = os.path.join(path, filename)
            if os.path.exists(candidate):
                return candidate
        return ""


def _unix_editor():
    macos_textedit = "/Applications/TextEdit.app/Contents/MacOS/TextEdit"

    if os.path.exists(macos_textedit):
        return macos_textedit
    else:
        return "gedit"


class UnixPlatform(Platform):
    """A platform module suitable for UNIX systems"""
    def __init__(self, basepath):
        if not basepath:
            basepath = os.path.join(self.default_dir(),
                                    ".chirp")

        Path(basepath).mkdir(exist_ok=True)
        super().__init__(str(basepath))

    def default_dir(self):
        return str(Path.home())

    def filter_filename(self, filename):
        return filename.replace("/", "")

    def open_text_file(self, path):
        pid1 = os.fork()
        if pid1 == 0:
            pid2 = os.fork()
            if pid2 == 0:
                editor = _unix_editor()
                LOG.debug("calling `%s %s'" % (editor, path))
                os.execlp(editor, editor, path)
            else:
                sys.exit(0)
        else:
            os.waitpid(pid1, 0)
            LOG.debug("Exec child exited")

    def open_html_file(self, path):
        os.system("firefox '%s'" % path)

    def os_version_string(self):
        try:
            issue = open("/etc/issue.net", "r")
            ver = issue.read().strip().replace("\r", "").replace("\n", "")[:64]
            issue.close()
            ver = "%s - %s" % (os.uname()[0], ver)
        except Exception:
            ver = " ".join(os.uname())

        return ver


class Win32Platform(Platform):
    """A platform module suitable for Windows systems"""
    def __init__(self, basepath=None):
        if not basepath:
            appdata = os.getenv("APPDATA")
            if not appdata:
                appdata = "C:\\"
            basepath = os.path.abspath(os.path.join(appdata, "CHIRP"))

        if not os.path.isdir(basepath):
            try:
                os.mkdir(basepath)
            except FileExistsError:
                pass

        Platform.__init__(self, basepath)

    def default_dir(self):
        return os.path.abspath(os.path.join(os.getenv("USERPROFILE"),
                                            "Desktop"))

    def filter_filename(self, filename):
        for char in "/\\:*?\"<>|":
            filename = filename.replace(char, "")

        return filename

    def open_text_file(self, path):
        Popen(["notepad", path])
        return

    def open_html_file(self, path):
        os.system("explorer %s" % path)

    def os_version_string(self):
        import win32api

        vers = {4: "Win2k",
                5: "WinXP",
                6: "WinVista/7",
                }

        (pform, sub, build, _, _) = win32api.GetVersionEx()

        return vers.get(pform,
                        "Win32 (Unknown %i.%i:%i)" % (pform, sub, build))


def _get_platform(basepath):
    if os.name == "nt":
        return Win32Platform(basepath)
    else:
        return UnixPlatform(basepath)


PLATFORM = None


def get_platform(basepath=None):
    """Return the platform singleton"""
    global PLATFORM

    if not PLATFORM:
        PLATFORM = _get_platform(basepath)

    return PLATFORM


def _do_test():
    __pform = get_platform()

    print("Config dir: %s" % __pform.config_dir())
    print("Default dir: %s" % __pform.default_dir())
    print("Log file (foo): %s" % __pform.log_file("foo"))
    print("OS Version: %s" % __pform.os_version_string())
    # __pform.open_text_file("d-rats.py")


if __name__ == "__main__":
    _do_test()



# Copyright 2014 Jens Jensen <af5mi@yahoo.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp.drivers import yaesu_clone, ft7800
from chirp import chirp_common, directory, bitwise

MEM_FORMAT = """
#seekto 0x018A;
struct {
    u16 in_use;
} bank_used[24];

#seekto 0x0214;
u16  banksoff1;
#seekto 0x0294;
u16  banksoff2;

#seekto 0x097A;
struct {
  u8 name[6];
} bank_names[24];

#seekto 0x0C0A;
struct {
  u16 channels[100];
} banks[24];

#seekto 0x0168;
struct {
  u8 used:1,
     unknown1:1,
     mode:1,
     unknown2:2,
     duplex:3;
  bbcd freq[3];
  u8 clockshift:1,
     tune_step:3,
     unknown5:1,
     tmode:3;
  bbcd split[3];
  u8 power:2,
     tone:6;
  u8 unknown6:1,
     dtcs:7;
  u8 unknown7[2];
  u8 offset;
  u8 unknown9[3];
} memory [200];

#seekto 0x0F28;
struct {
  char name[6];
  u8 enabled:1,
     unknown1:7;
  u8 used:1,
     unknown2:7;
} names[200];

#seekto 0x1768;
struct {
  u8 skip3:2,
     skip2:2,
     skip1:2,
     skip0:2;
} flags[50];
"""


@directory.register
class VX170Radio(ft7800.FTx800Radio):
    """Yaesu VX-170"""
    MODEL = "VX-170"
    _model = "AH022"
    _memsize = 6057
    _block_lengths = [8, 6048, 1]
    _block_size = 32

    POWER_LEVELS_VHF = [chirp_common.PowerLevel("Hi", watts=5.00),
                        chirp_common.PowerLevel("Med", watts=2.00),
                        chirp_common.PowerLevel("Lo", watts=0.50)]

    MODES = ["FM", "NFM"]
    TMODES = ["", "Tone", "TSQL", "DTCS"]

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.pre_download = _(
            "1. Turn radio off.\n"
            "2. Connect cable to MIC/SP jack.\n"
            "3. Press and hold in the [moni] key while turning the radio"
            " on.\n"
            "4. Select CLONE in menu, then press F. Radio restarts in clone"
            " mode.\n"
            "     (\"CLONE\" will appear on the display).\n"
            "5. <b>After clicking OK</b>, briefly hold [PTT] key to send"
            " image.\n"
            "    (\"-TX-\" will appear on the LCD). \n")
        rp.pre_upload = _(
            "1. Turn radio off.\n"
            "3. Press and hold in the [moni] key while turning the radio"
            " on.\n"
            "4. Select CLONE in menu, then press F. Radio restarts in clone"
            " mode.\n"
            "     (\"CLONE\" will appear on the display).\n"
            "5. Press the [moni] key (\"-RX-\" will appear on the LCD).\n")
        return rp

    def _checksums(self):
        return [yaesu_clone.YaesuChecksum(0x0000, self._memsize - 2)]

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_features(self):
        rf = super(VX170Radio, self).get_features()
        rf.has_bank = False
        rf.has_bank_names = False
        rf.valid_modes = self.MODES
        rf.memory_bounds = (1, 200)
        rf.valid_bands = [(137000000, 174000000)]
        return rf



# Copyright 2014 Tom Hayward <tom@tomh.us>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging

from chirp import chirp_common, directory, memmap, errors, util
from chirp import bitwise
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueInteger, RadioSettingValueList, \
    RadioSettingValueBoolean, RadioSettingValueString, \
    RadioSettings

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
#seekto 0x0184;
struct {
  u8 unknown1:4,
     sql:4;              // squelch level
  u8 unknown2;
  u8 obeep:1,            // open beep
     dw_off:1,           // dual watch (inverted)
     kbeep:1,            // key beep
     rbeep:1,            // roger beep
     unknown3:2,
     ctdcsb:1,           // ct/dcs busy lock
     unknown4:1;
  u8 alarm:1,            // alarm key
     unknown5:1,
     aliasen_off:1,      // alias enable (inverted)
     save:1,             // battery save
     unknown6:2,
     mrcha:1,            // mr/cha
     vfomr:1;            // vfo/mr
  u8 keylock_off:1,      // key lock (inverted)
     txstop_off:1,       // tx stop (inverted)
     scanm:1,            // scan key/mode
     vir:1,              // vox inhibit on receive
     keylockm:2,         // key lock mode
     lamp:2;             // backlight
  u8 opendis:2,          // open display
     fmen_off:1,         // fm enable (inverted)
     unknown7:1,
     fmscan_off:1,       // fm scan (inverted)
     fmdw:1,             // fm dual watch
     unknown8:2;
  u8 step:4,             // step
     vol:4;              // volume
  u8 apo:4,              // auto power off
     tot:4;              // time out timer
  u8 unknown9;
  u8 voxdt:4,            // vox delay time
     voxgain:4;          // vox gain
  u8 unknown10;
  u8 unknown11;
  u8 unknown12:3,
     lptime:5;           // long press time
  u8 keyp2long:4,        // p2 key long press
     keyp2short:4;       // p2 key short press
  u8 keyp1long:4,        // p1 key long press
     keyp1short:4;       // p1 key short press
  u8 keyp3long:4,        // p3 key long press
     keyp3short:4;       // p3 key short press
  u8 unknown13;
  u8 menuen:1,           // menu enable
     absel:1,            // a/b select
     unknown14:2,
     keymshort:4;        // m key short press
  u8 unknown15:4,
     dtmfst:1,           // dtmf sidetone
     ackdecode:1,        // ack decode
     monitor:2;          // monitor
  u8 unknown16:3,
     reset:1,            // reset enable
     unknown17:1,
     keypadmic_off:1,    // keypad mic (inverted)
     unknown18:2;
  u8 unknown19;
  u8 unknown20:3,
     dtmftime:5;         // dtmf digit time
  u8 unknown21:3,
     dtmfspace:5;        // dtmf digit space time
  u8 unknown22:2,
     dtmfdelay:6;        // dtmf first digit delay
  u8 unknown23:1,
     dtmfpretime:7;      // dtmf pretime
  u8 unknown24:2,
     dtmfdelay2:6;       // dtmf * and # digit delay
  u8 unknown25:3,
     smfont_off:1,       // small font (inverted)
     unknown26:4;
} settings;

#seekto 0x01cd;
struct {
  u8 rssi136;            // squelch base level (vhf)
  u8 unknown0x01ce;
  u8 rssi400;            // squelch base level (uhf)
} service;

#seekto 0x0900;
struct {
  char user1[7];         // user message 1
  char unknown0x0907;
  char unknown0x0908[8];
  char unknown0x0910[8];
  char system[7];        // system message
  char unknown0x091F;
  char user2[7];         // user message 2
  char unknown0x0927;
} messages;

struct channel {
  bbcd rx_freq[4];
  bbcd tx_freq[4];
  u8 rx_tone;
  u8 rx_tmode_extra:6,
     rx_tmode:2;
  u8 tx_tone;
  u8 tx_tmode_extra:6,
     tx_tmode:2;
  u8 unknown5;
  u8 pttidoff:1,
     dtmfoff:1,
     %(unknownormode)s,
     tailcut:1,
     aliasop:1,
     talkaroundoff:1,
     voxoff:1,
     skip:1;
  u8 %(modeorpower)s,
     reverseoff:1,
     blckoff:1,
     unknown7:1,
     apro:3;
  u8 unknown8;
};

struct name {
    char name[7];
    u8 pad;
};

#seekto 0x%(chanstart)x;
struct channel default[%(defaults)i];
struct channel memory[199];

#seekto 0x%(namestart)x;
struct name defaultname[%(defaults)i];
struct name name[199];
"""


APO_LIST = ["OFF", "10M", "20M", "30M", "40M", "50M", "60M", "90M",
            "2H", "4H", "6H", "8H", "10H", "12H", "14H", "16H"]
SQL_LIST = ["%s" % x for x in range(0, 10)]
SCANM_LIST = ["CO", "TO"]
TOT_LIST = ["OFF"] + ["%s seconds" % x for x in range(10, 130, 10)]
_STEP_LIST = [2.5, 5., 6.25, 10., 12.5, 25.]
STEP_LIST = ["{} kHz".format(x) for x in _STEP_LIST]
MONITOR_LIST = ["CTC/DCS", "DTMF", "CTC/DCS and DTMF", "CTC/DCS or DTMF"]
VFOMR_LIST = ["MR", "VFO"]
MRCHA_LIST = ["MR CHA", "Freq. MR"]
VOL_LIST = ["OFF"] + ["%s" % x for x in range(1, 16)]
OPENDIS_LIST = ["All", "Lease Time", "User-defined", "Leixen"]
LAMP_LIST = ["OFF", "KEY", "CONT"]
KEYLOCKM_LIST = ["K+S", "PTT", "KEY", "ALL"]
ABSEL_LIST = ["B Channel",  "A Channel"]
VOXGAIN_LIST = ["%s" % x for x in range(1, 9)]
VOXDT_LIST = ["%s seconds" % x for x in range(1, 5)]
DTMFTIME_LIST = ["%i milliseconds" % x for x in range(50, 210, 10)]
DTMFDELAY_LIST = ["%i milliseconds" % x for x in range(0, 550, 50)]
DTMFPRETIME_LIST = ["%i milliseconds" % x for x in range(100, 1100, 100)]
DTMFDELAY2_LIST = ["%i milliseconds" % x for x in range(0, 450, 50)]

LPTIME_LIST = ["%i milliseconds" % x for x in range(500, 2600, 100)]
PFKEYLONG_LIST = ["OFF",
                  "FM",
                  "Monitor Momentary",
                  "Monitor Lock",
                  "SQ Off Momentary",
                  "Mute",
                  "SCAN",
                  "TX Power",
                  "EMG",
                  "VFO/MR",
                  "DTMF",
                  "CALL",
                  "Transmit 1750 Hz",
                  "A/B",
                  "Talk Around",
                  "Reverse"
                  ]

PFKEYSHORT_LIST = ["OFF",
                   "FM",
                   "BandChange",
                   "Time",
                   "Monitor Lock",
                   "Mute",
                   "SCAN",
                   "TX Power",
                   "EMG",
                   "VFO/MR",
                   "DTMF",
                   "CALL",
                   "Transmit 1750 Hz",
                   "A/B",
                   "Talk Around",
                   "Reverse"
                   ]

MODES = ["NFM", "FM"]
WTFTONES = tuple(float(x) for x in range(56, 64))
TONES = tuple(sorted(WTFTONES + chirp_common.TONES))
DTCS_CODES = tuple(sorted((17, 50, 645) + chirp_common.DTCS_CODES))
TMODES = ["", "Tone", "DTCS", "DTCS"]


def _image_ident_from_data(data):
    return data[0x168:0x178]


def _image_ident_from_image(radio):
    return _image_ident_from_data(radio.get_mmap())


def checksum(frame):
    x = 0
    for b in frame:
        x ^= b
    return x & 0xFF


def make_frame(cmd, addr, data=b""):
    payload = struct.pack(">H", addr) + data
    header = struct.pack(">cB", cmd, len(payload))
    frame = header + payload
    return frame + bytes([checksum(frame)])


def send(radio, frame):
    # LOG.debug("%04i P>R: %s" %
    #           (len(frame),
    #            util.hexprint(frame).replace("\n", "\n          ")))
    try:
        radio.pipe.write(frame)
    except Exception as e:
        raise errors.RadioError("Failed to communicate with radio: %s" % e)


def recv(radio, readdata=True):
    hdr = radio.pipe.read(4)
    # LOG.debug("%04i P<R: %s" %
    #           (len(hdr), util.hexprint(hdr).replace("\n", "\n          ")))
    if hdr == b"\x09\x00\x09":
        raise errors.RadioError("Radio rejected command.")
    cmd, length, addr = struct.unpack(">BBH", hdr)
    length -= 2
    if readdata:
        data = radio.pipe.read(length)
        # LOG.debug("     P<R: %s" %
        #           util.hexprint(hdr + data).replace("\n", "\n          "))
        if len(data) != length:
            raise errors.RadioError("Radio sent %i bytes (expected %i)" % (
                len(data), length))
        chk = radio.pipe.read(1)
    else:
        data = b""
    return addr, data


def do_ident(radio):
    send(radio, b"\x02\x06LEIXEN\x17")
    ident = radio.pipe.read(9)
    LOG.debug("     P<R: %s" %
              util.hexprint(ident).replace("\n", "\n          "))
    if ident != b"\x06\x06leixen\x13":
        raise errors.RadioError("Radio refused program mode")
    radio.pipe.write(b"\x06\x00\x06")
    ack = radio.pipe.read(3)
    if ack != b"\x06\x00\x06":
        raise errors.RadioError("Radio did not ack.")


def do_download(radio):
    # Ident should have already been done by the detect_from_serial()

    data = b""
    data += b"\xFF" * (0 - len(data))
    for addr in range(0, radio._memsize, 0x10):
        send(radio, make_frame(b"R", addr, b'\x10'))
        _addr, _data = recv(radio)
        if _addr != addr:
            raise errors.RadioError("Radio sent unexpected address")
        data += _data

        status = chirp_common.Status()
        status.cur = addr
        status.max = radio._memsize
        status.msg = "Cloning from radio"
        radio.status_fn(status)

    finish(radio)

    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    _ranges = [(0x0d00, 0x2000)]

    image_ident = _image_ident_from_image(radio)
    if image_ident.startswith(radio._file_ident) and \
       radio._model_ident in image_ident:
        _ranges = radio._ranges

    do_ident(radio)

    for start, end in _ranges:
        LOG.debug('Uploading range 0x%04X - 0x%04X' % (start, end))
        for addr in range(start, end, 0x10):
            frame = make_frame(b"W", addr, radio._mmap[addr:addr + 0x10])
            send(radio, frame)
            # LOG.debug("     P<R: %s" %
            #           util.hexprint(frame).replace("\n", "\n          "))
            radio.pipe.write(b"\x06\x00\x06")
            ack = radio.pipe.read(3)
            if ack != b"\x06\x00\x06":
                raise errors.RadioError("Radio refused block at %04x" % addr)

            status = chirp_common.Status()
            status.cur = addr
            status.max = radio._memsize
            status.msg = "Cloning to radio"
            radio.status_fn(status)

    finish(radio)


def finish(radio):
    send(radio, b"\x64\x01\x6F\x0A")
    ack = radio.pipe.read(8)


@directory.register
class LeixenVV898Radio(chirp_common.CloneModeRadio):

    """Leixen VV-898"""
    VENDOR = "Leixen"
    MODEL = "VV-898"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False

    _file_ident = b"Leixen"
    _model_ident = b'LX-\x89\x85\x63'

    _memsize = 0x2000
    _ranges = [
        (0x0000, 0x013f),
        (0x0148, 0x0167),
        (0x0184, 0x018f),
        (0x0190, 0x01cf),
        (0x0900, 0x090f),
        (0x0920, 0x0927),
        (0x0d00, 0x2000),
    ]

    _mem_formatter = {'unknownormode': 'unknown6:1',
                      'modeorpower': 'mode:1, power:1',
                      'chanstart': 0x0D00,
                      'namestart': 0x19B0,
                      'defaults': 3}
    _power_levels = [chirp_common.PowerLevel("Low", watts=4),
                     chirp_common.PowerLevel("High", watts=10)]

    @classmethod
    def detect_from_serial(cls, pipe):
        radio = cls(pipe)
        do_ident(radio)
        send(radio, make_frame(b"R", 0x0168, b'\x10'))
        _addr, _data = recv(radio)
        ident = _data[8:14]
        LOG.debug('Got ident from radio:\n%s' % util.hexprint(ident))
        for rclass in [cls] + cls.detected_models():
            if ident == rclass._model_ident:
                return rclass
        # Reset the radio if we didn't find a match
        finish(radio)
        raise errors.RadioError('Unable to detect a supported model')

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_cross = True
        rf.has_bank = False
        rf.has_tuning_step = False
        rf.can_odd_split = True
        rf.has_rx_dtcs = True
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_modes = MODES
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->",
            "->DTCS",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "DTCS->DTCS"]
        rf.valid_characters = chirp_common.CHARSET_ASCII
        rf.valid_name_length = 7
        rf.valid_power_levels = self._power_levels
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.valid_skips = ["", "S"]
        rf.valid_tuning_steps = _STEP_LIST
        rf.valid_bands = [(136000000, 174000000),
                          (400000000, 480000000)]
        rf.valid_tones = TONES
        rf.valid_dtcs_codes = DTCS_CODES
        rf.memory_bounds = (1, 199)
        return rf

    def sync_in(self):
        try:
            self._mmap = do_download(self)
        except Exception as e:
            finish(self)
            raise errors.RadioError("Failed to download from radio: %s" % e)
        self.process_mmap()

    def process_mmap(self):
        self._memobj = bitwise.parse(
            MEM_FORMAT % self._mem_formatter, self._mmap)

    def sync_out(self):
        try:
            do_upload(self)
        except errors.RadioError:
            finish(self)
            raise
        except Exception as e:
            raise errors.RadioError("Failed to upload to radio: %s" % e)

    def get_raw_memory(self, number):
        name, mem = self._get_memobjs(number)
        return repr(name) + repr(mem)

    def _get_tone(self, mem, _mem):
        rx_tone = tx_tone = None

        tx_tmode = TMODES[_mem.tx_tmode]
        rx_tmode = TMODES[_mem.rx_tmode]

        if tx_tmode == "Tone":
            tx_tone = TONES[_mem.tx_tone - 1]
        elif tx_tmode == "DTCS":
            tx_tone = DTCS_CODES[_mem.tx_tone - 1]

        if rx_tmode == "Tone":
            rx_tone = TONES[_mem.rx_tone - 1]
        elif rx_tmode == "DTCS":
            rx_tone = DTCS_CODES[_mem.rx_tone - 1]

        tx_pol = _mem.tx_tmode == 0x03 and "R" or "N"
        rx_pol = _mem.rx_tmode == 0x03 and "R" or "N"

        chirp_common.split_tone_decode(mem, (tx_tmode, tx_tone, tx_pol),
                                       (rx_tmode, rx_tone, rx_pol))

    def _is_txinh(self, _mem):
        raw_tx = b""
        for i in range(0, 4):
            raw_tx += _mem.tx_freq[i].get_raw()
        return raw_tx == b"\xFF\xFF\xFF\xFF"

    def _get_memobjs(self, number):
        _mem = self._memobj.memory[number - 1]
        _name = self._memobj.name[number - 1]
        return _mem, _name

    def get_memory(self, number):
        _mem, _name = self._get_memobjs(number)

        mem = chirp_common.Memory()
        mem.number = number

        if _mem.get_raw()[:4] == b"\xFF\xFF\xFF\xFF":
            mem.empty = True
            return mem

        mem.freq = int(_mem.rx_freq) * 10

        if self._is_txinh(_mem):
            mem.duplex = "off"
            mem.offset = 0
        elif int(_mem.rx_freq) == int(_mem.tx_freq):
            mem.duplex = ""
            mem.offset = 0
        elif abs(int(_mem.rx_freq) * 10 - int(_mem.tx_freq) * 10) > 70000000:
            mem.duplex = "split"
            mem.offset = int(_mem.tx_freq) * 10
        else:
            mem.duplex = int(_mem.rx_freq) > int(_mem.tx_freq) and "-" or "+"
            mem.offset = abs(int(_mem.rx_freq) - int(_mem.tx_freq)) * 10

        mem.name = str(_name.name).rstrip()

        self._get_tone(mem, _mem)
        mem.mode = MODES[_mem.mode]
        powerindex = _mem.power if _mem.power < len(self._power_levels) else -1
        mem.power = self._power_levels[powerindex]
        mem.skip = _mem.skip and "S" or ""

        mem.extra = RadioSettingGroup("Extra", "extra")

        opts = ["On", "Off"]
        rs = RadioSetting("blckoff", "Busy Channel Lockout",
                          RadioSettingValueList(
                              opts, opts[_mem.blckoff]))
        mem.extra.append(rs)
        opts = ["Off", "On"]
        rs = RadioSetting("tailcut", "Squelch Tail Elimination",
                          RadioSettingValueList(
                              opts, opts[_mem.tailcut]))
        mem.extra.append(rs)
        apro = _mem.apro if _mem.apro < 0x5 else 0
        opts = ["Off", "Compander", "Scrambler", "TX Scrambler",
                "RX Scrambler"]
        rs = RadioSetting("apro", "Audio Processing",
                          RadioSettingValueList(
                              opts, opts[apro]))
        mem.extra.append(rs)
        opts = ["On", "Off"]
        rs = RadioSetting("voxoff", "VOX",
                          RadioSettingValueList(
                              opts, opts[_mem.voxoff]))
        mem.extra.append(rs)
        opts = ["On", "Off"]
        rs = RadioSetting("pttidoff", "PTT ID",
                          RadioSettingValueList(
                              opts, opts[_mem.pttidoff]))
        mem.extra.append(rs)
        opts = ["On", "Off"]
        rs = RadioSetting("dtmfoff", "DTMF",
                          RadioSettingValueList(
                              opts, opts[_mem.dtmfoff]))
        mem.extra.append(rs)
        opts = ["Name", "Frequency"]
        aliasop = RadioSetting("aliasop", "Display",
                               RadioSettingValueList(
                                   opts, opts[_mem.aliasop]))
        mem.extra.append(aliasop)
        opts = ["On", "Off"]
        rs = RadioSetting("reverseoff", "Reverse Frequency",
                          RadioSettingValueList(
                              opts, opts[_mem.reverseoff]))
        mem.extra.append(rs)
        opts = ["On", "Off"]
        rs = RadioSetting("talkaroundoff", "Talk Around",
                          RadioSettingValueList(
                              opts, opts[_mem.talkaroundoff]))
        mem.extra.append(rs)

        return mem

    def _set_tone(self, mem, _mem):
        ((txmode, txtone, txpol),
         (rxmode, rxtone, rxpol)) = chirp_common.split_tone_encode(mem)

        _mem.tx_tmode = TMODES.index(txmode)
        _mem.rx_tmode = TMODES.index(rxmode)
        if txmode == "Tone":
            _mem.tx_tone = TONES.index(txtone) + 1
        elif txmode == "DTCS":
            _mem.tx_tmode = txpol == "R" and 0x03 or 0x02
            _mem.tx_tone = DTCS_CODES.index(txtone) + 1
        if rxmode == "Tone":
            _mem.rx_tone = TONES.index(rxtone) + 1
        elif rxmode == "DTCS":
            _mem.rx_tmode = rxpol == "R" and 0x03 or 0x02
            _mem.rx_tone = DTCS_CODES.index(rxtone) + 1

    def set_memory(self, mem):
        _mem, _name = self._get_memobjs(mem.number)

        if mem.empty:
            _mem.set_raw(b"\xFF" * 16)
            return
        elif _mem.get_raw() == (b"\xFF" * 16):
            _mem.set_raw(b"\xFF" * 8 + b"\xFF\x00\xFF\x00\xFF\xFE\xF0\xFC")

        _mem.rx_freq = mem.freq / 10

        if mem.duplex == "off":
            for i in range(0, 4):
                _mem.tx_freq[i].set_raw(b"\xFF")
        elif mem.duplex == "split":
            _mem.tx_freq = mem.offset / 10
        elif mem.duplex == "+":
            _mem.tx_freq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.tx_freq = (mem.freq - mem.offset) / 10
        else:
            _mem.tx_freq = mem.freq / 10

        self._set_tone(mem, _mem)

        _mem.power = mem.power and self._power_levels.index(mem.power) or 0
        _mem.mode = MODES.index(mem.mode)
        _mem.skip = mem.skip == "S"
        _name.name = mem.name.ljust(7)

        # autoset display to name if filled, else show frequency
        if mem.extra:
            # mem.extra only seems to be populated when called from edit panel
            aliasop = mem.extra["aliasop"]
        else:
            aliasop = None
        if mem.name:
            _mem.aliasop = False
        else:
            _mem.aliasop = True

        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)

    def _get_settings(self):
        _settings = self._memobj.settings
        _service = self._memobj.service
        _msg = self._memobj.messages
        cfg_grp = RadioSettingGroup("cfg_grp", "Basic Settings")
        adv_grp = RadioSettingGroup("adv_grp", "Advanced Settings")
        key_grp = RadioSettingGroup("key_grp", "Key Assignment")
        group = RadioSettings(cfg_grp, adv_grp, key_grp)

        #
        # Basic Settings
        #
        rs = RadioSetting("apo", "Auto Power Off",
                          RadioSettingValueList(
                              APO_LIST, APO_LIST[_settings.apo]))
        cfg_grp.append(rs)
        rs = RadioSetting("sql", "Squelch Level",
                          RadioSettingValueList(
                              SQL_LIST, SQL_LIST[_settings.sql]))
        cfg_grp.append(rs)
        rs = RadioSetting("scanm", "Scan Mode",
                          RadioSettingValueList(
                              SCANM_LIST, SCANM_LIST[_settings.scanm]))
        cfg_grp.append(rs)
        rs = RadioSetting("tot", "Time Out Timer",
                          RadioSettingValueList(
                              TOT_LIST, TOT_LIST[_settings.tot]))
        cfg_grp.append(rs)
        rs = RadioSetting("step", "Step",
                          RadioSettingValueList(
                              STEP_LIST, STEP_LIST[_settings.step]))
        cfg_grp.append(rs)
        rs = RadioSetting("monitor", "Monitor",
                          RadioSettingValueList(
                              MONITOR_LIST, MONITOR_LIST[_settings.monitor]))
        cfg_grp.append(rs)
        rs = RadioSetting("vfomr", "VFO/MR",
                          RadioSettingValueList(
                              VFOMR_LIST, VFOMR_LIST[_settings.vfomr]))
        cfg_grp.append(rs)
        rs = RadioSetting("mrcha", "MR/CHA",
                          RadioSettingValueList(
                              MRCHA_LIST, MRCHA_LIST[_settings.mrcha]))
        cfg_grp.append(rs)
        rs = RadioSetting("vol", "Volume",
                          RadioSettingValueList(
                              VOL_LIST, VOL_LIST[_settings.vol]))
        cfg_grp.append(rs)
        rs = RadioSetting("opendis", "Open Display",
                          RadioSettingValueList(
                              OPENDIS_LIST, OPENDIS_LIST[_settings.opendis]))
        cfg_grp.append(rs)

        def _filter(name):
            filtered = ""
            for char in str(name):
                if char in chirp_common.CHARSET_ASCII:
                    filtered += char
                else:
                    filtered += " "
            LOG.debug("Filtered: %s" % filtered)
            return filtered

        rs = RadioSetting("messages.user1", "User-defined Message 1",
                          RadioSettingValueString(0, 7, _filter(_msg.user1)))
        cfg_grp.append(rs)
        rs = RadioSetting("messages.user2", "User-defined Message 2",
                          RadioSettingValueString(0, 7, _filter(_msg.user2)))
        cfg_grp.append(rs)

        val = RadioSettingValueString(0, 7, _filter(_msg.system))
        val.set_mutable(False)
        rs = RadioSetting("messages.system", "System Message", val)
        cfg_grp.append(rs)

        rs = RadioSetting("lamp", "Backlight",
                          RadioSettingValueList(
                              LAMP_LIST, LAMP_LIST[_settings.lamp]))
        cfg_grp.append(rs)
        rs = RadioSetting("keylockm", "Key Lock Mode",
                          RadioSettingValueList(
                              KEYLOCKM_LIST,
                              KEYLOCKM_LIST[_settings.keylockm]))
        cfg_grp.append(rs)
        rs = RadioSetting("absel", "A/B Select",
                          RadioSettingValueList(ABSEL_LIST,
                                                ABSEL_LIST[_settings.absel]))
        cfg_grp.append(rs)

        rs = RadioSetting("obeep", "Open Beep",
                          RadioSettingValueBoolean(_settings.obeep))
        cfg_grp.append(rs)
        rs = RadioSetting("rbeep", "Roger Beep",
                          RadioSettingValueBoolean(_settings.rbeep))
        cfg_grp.append(rs)
        rs = RadioSetting("keylock_off", "Key Lock",
                          RadioSettingValueBoolean(not _settings.keylock_off))
        cfg_grp.append(rs)
        rs = RadioSetting("ctdcsb", "CT/DCS Busy Lock",
                          RadioSettingValueBoolean(_settings.ctdcsb))
        cfg_grp.append(rs)
        rs = RadioSetting("alarm", "Alarm Key",
                          RadioSettingValueBoolean(_settings.alarm))
        cfg_grp.append(rs)
        rs = RadioSetting("save", "Battery Save",
                          RadioSettingValueBoolean(_settings.save))
        cfg_grp.append(rs)
        rs = RadioSetting("kbeep", "Key Beep",
                          RadioSettingValueBoolean(_settings.kbeep))
        cfg_grp.append(rs)
        rs = RadioSetting("reset", "Reset Enable",
                          RadioSettingValueBoolean(_settings.reset))
        cfg_grp.append(rs)
        rs = RadioSetting("smfont_off", "Small Font",
                          RadioSettingValueBoolean(not _settings.smfont_off))
        cfg_grp.append(rs)
        rs = RadioSetting("aliasen_off", "Alias Enable",
                          RadioSettingValueBoolean(not _settings.aliasen_off))
        cfg_grp.append(rs)
        rs = RadioSetting("txstop_off", "TX Stop",
                          RadioSettingValueBoolean(not _settings.txstop_off))
        cfg_grp.append(rs)
        rs = RadioSetting("dw_off", "Dual Watch",
                          RadioSettingValueBoolean(not _settings.dw_off))
        cfg_grp.append(rs)
        rs = RadioSetting("fmen_off", "FM Enable",
                          RadioSettingValueBoolean(not _settings.fmen_off))
        cfg_grp.append(rs)
        rs = RadioSetting("fmdw", "FM Dual Watch",
                          RadioSettingValueBoolean(_settings.fmdw))
        cfg_grp.append(rs)
        rs = RadioSetting("fmscan_off", "FM Scan",
                          RadioSettingValueBoolean(
                              not _settings.fmscan_off))
        cfg_grp.append(rs)
        rs = RadioSetting("keypadmic_off", "Keypad MIC",
                          RadioSettingValueBoolean(
                              not _settings.keypadmic_off))
        cfg_grp.append(rs)
        rs = RadioSetting("voxgain", "VOX Gain",
                          RadioSettingValueList(
                              VOXGAIN_LIST, VOXGAIN_LIST[_settings.voxgain]))
        cfg_grp.append(rs)
        rs = RadioSetting("voxdt", "VOX Delay Time",
                          RadioSettingValueList(
                              VOXDT_LIST, VOXDT_LIST[_settings.voxdt]))
        cfg_grp.append(rs)
        rs = RadioSetting("vir", "VOX Inhibit on Receive",
                          RadioSettingValueBoolean(_settings.vir))
        cfg_grp.append(rs)

        #
        # Advanced Settings
        #
        val = (_settings.dtmftime) - 5
        rs = RadioSetting("dtmftime", "DTMF Digit Time",
                          RadioSettingValueList(
                              DTMFTIME_LIST, DTMFTIME_LIST[val]))
        adv_grp.append(rs)
        val = (_settings.dtmfspace) - 5
        rs = RadioSetting("dtmfspace", "DTMF Digit Space Time",
                          RadioSettingValueList(
                              DTMFTIME_LIST, DTMFTIME_LIST[val]))
        adv_grp.append(rs)
        val = (_settings.dtmfdelay) // 5
        rs = RadioSetting("dtmfdelay", "DTMF 1st Digit Delay",
                          RadioSettingValueList(
                              DTMFDELAY_LIST, DTMFDELAY_LIST[val]))
        adv_grp.append(rs)
        val = (_settings.dtmfpretime) // 10 - 1
        rs = RadioSetting("dtmfpretime", "DTMF Pretime",
                          RadioSettingValueList(
                              DTMFPRETIME_LIST, DTMFPRETIME_LIST[val]))
        adv_grp.append(rs)
        val = (_settings.dtmfdelay2) // 5
        rs = RadioSetting("dtmfdelay2", "DTMF * and # Digit Delay",
                          RadioSettingValueList(
                              DTMFDELAY2_LIST, DTMFDELAY2_LIST[val]))
        adv_grp.append(rs)
        rs = RadioSetting("ackdecode", "ACK Decode",
                          RadioSettingValueBoolean(_settings.ackdecode))
        adv_grp.append(rs)
        rs = RadioSetting("dtmfst", "DTMF Sidetone",
                          RadioSettingValueBoolean(_settings.dtmfst))
        adv_grp.append(rs)

        rs = RadioSetting("service.rssi400", "Squelch Base Level (UHF)",
                          RadioSettingValueInteger(0, 255, _service.rssi400))
        adv_grp.append(rs)
        rs = RadioSetting("service.rssi136", "Squelch Base Level (VHF)",
                          RadioSettingValueInteger(0, 255, _service.rssi136))
        adv_grp.append(rs)

        #
        # Key Settings
        #
        val = (_settings.lptime) - 5
        rs = RadioSetting("lptime", "Long Press Time",
                          RadioSettingValueList(
                              LPTIME_LIST, LPTIME_LIST[val]))
        key_grp.append(rs)
        rs = RadioSetting("keyp1long", "P1 Long Key",
                          RadioSettingValueList(
                              PFKEYLONG_LIST,
                              PFKEYLONG_LIST[_settings.keyp1long]))
        key_grp.append(rs)
        rs = RadioSetting("keyp1short", "P1 Short Key",
                          RadioSettingValueList(
                              PFKEYSHORT_LIST,
                              PFKEYSHORT_LIST[_settings.keyp1short]))
        key_grp.append(rs)
        rs = RadioSetting("keyp2long", "P2 Long Key",
                          RadioSettingValueList(
                              PFKEYLONG_LIST,
                              PFKEYLONG_LIST[_settings.keyp2long]))
        key_grp.append(rs)
        rs = RadioSetting("keyp2short", "P2 Short Key",
                          RadioSettingValueList(
                              PFKEYSHORT_LIST,
                              PFKEYSHORT_LIST[_settings.keyp2short]))
        key_grp.append(rs)
        rs = RadioSetting("keyp3long", "P3 Long Key",
                          RadioSettingValueList(
                              PFKEYLONG_LIST,
                              PFKEYLONG_LIST[_settings.keyp3long]))
        key_grp.append(rs)
        rs = RadioSetting("keyp3short", "P3 Short Key",
                          RadioSettingValueList(
                              PFKEYSHORT_LIST,
                              PFKEYSHORT_LIST[_settings.keyp3short]))
        key_grp.append(rs)

        val = RadioSettingValueList(PFKEYSHORT_LIST,
                                    PFKEYSHORT_LIST[_settings.keymshort])
        val.set_mutable(_settings.menuen == 0)
        rs = RadioSetting("keymshort", "M Short Key", val)
        key_grp.append(rs)
        val = RadioSettingValueBoolean(_settings.menuen)
        rs = RadioSetting("menuen", "Menu Enable", val)
        key_grp.append(rs)

        return group

    def get_settings(self):
        try:
            return self._get_settings()
        except:
            import traceback
            LOG.error("Failed to parse settings: %s", traceback.format_exc())
            return None

    def set_settings(self, settings):
        _settings = self._memobj.settings
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    name = element.get_name()
                    if "." in name:
                        bits = name.split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            if "/" in bit:
                                bit, index = bit.split("/", 1)
                                index = int(index)
                                obj = getattr(obj, bit)[index]
                            else:
                                obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = _settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif setting == "keylock_off":
                        setattr(obj, setting, not int(element.value))
                    elif setting == "smfont_off":
                        setattr(obj, setting, not int(element.value))
                    elif setting == "aliasen_off":
                        setattr(obj, setting, not int(element.value))
                    elif setting == "txstop_off":
                        setattr(obj, setting, not int(element.value))
                    elif setting == "dw_off":
                        setattr(obj, setting, not int(element.value))
                    elif setting == "fmen_off":
                        setattr(obj, setting, not int(element.value))
                    elif setting == "fmscan_off":
                        setattr(obj, setting, not int(element.value))
                    elif setting == "keypadmic_off":
                        setattr(obj, setting, not int(element.value))
                    elif setting == "dtmftime":
                        setattr(obj, setting, int(element.value) + 5)
                    elif setting == "dtmfspace":
                        setattr(obj, setting, int(element.value) + 5)
                    elif setting == "dtmfdelay":
                        setattr(obj, setting, int(element.value) * 5)
                    elif setting == "dtmfpretime":
                        setattr(obj, setting, (int(element.value) + 1) * 10)
                    elif setting == "dtmfdelay2":
                        setattr(obj, setting, int(element.value) * 5)
                    elif setting == "lptime":
                        setattr(obj, setting, int(element.value) + 5)
                    else:
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    @classmethod
    def match_model(cls, filedata, filename):
        if filedata[0x168:0x170].startswith(cls._file_ident) and \
           filedata[0x170:0x178].startswith(cls._model_ident):
            return True
        else:
            return False


@directory.register
class JetstreamJT270MRadio(LeixenVV898Radio):

    """Jetstream JT270M"""
    VENDOR = "Jetstream"
    MODEL = "JT270M"
    ALIASES = []

    _file_ident = b"JET"
    _model_ident = b'LX-\x89\x85\x53'


class LT898UV(LeixenVV898Radio):
    VENDOR = "LUITON"
    MODEL = "LT-898UV"

    @classmethod
    def match_model(cls, filedata, filename):
        return False


@directory.register
class JetstreamJT270MHRadio(LeixenVV898Radio):

    """Jetstream JT270MH"""
    VENDOR = "Jetstream"
    MODEL = "JT270MH"

    _file_ident = b"Leixen"
    _model_ident = b'LX-\x89\x85\x85'
    _ranges = [(0x0C00, 0x2000)]
    _mem_formatter = {'unknownormode': 'mode:1',
                      'modeorpower': 'power:2',
                      'chanstart': 0x0C00,
                      'namestart': 0x1900,
                      'defaults': 6}
    _power_levels = [chirp_common.PowerLevel("Low", watts=5),
                     chirp_common.PowerLevel("Mid", watts=10),
                     chirp_common.PowerLevel("High", watts=25)]
    # Base radio has offset zero to distinguish from sub devices
    _offset = 0

    def get_features(self):
        rf = super(JetstreamJT270MHRadio, self).get_features()
        rf.has_sub_devices = self._offset == 0
        rf.memory_bounds = (1, 99)
        return rf

    def get_sub_devices(self):
        return [JetstreamJT270MHRadioA(self._mmap),
                JetstreamJT270MHRadioB(self._mmap)]

    def _get_memobjs(self, number):
        number = number * 2 - self._offset
        _mem = self._memobj.memory[number]
        _name = self._memobj.name[number]
        return _mem, _name


class JetstreamJT270MHRadioA(JetstreamJT270MHRadio):
    VARIANT = 'A Band'
    _offset = 1


class JetstreamJT270MHRadioB(JetstreamJT270MHRadio):
    VARIANT = 'B Band'
    _offset = 2


@directory.register
class LeixenVV898SRadio(LeixenVV898Radio):

    """Leixen VV-898S, also VV-898E which is identical"""
    VENDOR = "Leixen"
    MODEL = "VV-898S"

    _model_ident = b'LX-\x89\x85\x75'
    _mem_formatter = {'unknownormode': 'mode:1',
                      'modeorpower': 'power:2',
                      'chanstart': 0x0D00,
                      'namestart': 0x19B0,
                      'defaults': 3}
    _power_levels = [chirp_common.PowerLevel("Low", watts=5),
                     chirp_common.PowerLevel("Med", watts=10),
                     chirp_common.PowerLevel("High", watts=25)]


@directory.register
class VV898E(LeixenVV898SRadio):
    '''Leixen has called this radio both 898E and S historically, ident is
    identical'''
    VENDOR = "Leixen"
    MODEL = "VV-898E"

    @classmethod
    def match_model(cls, filedata, filename):
        return False


@directory.register
@directory.detected_by(LeixenVV898SRadio)
class VV898SDualBank(JetstreamJT270MHRadio):
    '''Newer VV898S 1.06+ firmware that features dual memory banks'''
    VENDOR = "Leixen"
    MODEL = "VV-898S"
    VARIANT = "Dual Bank"

    @classmethod
    def match_model(cls, filedata, filename):
        return False


@directory.register
@directory.detected_by(VV898E)
class VV898EDualBank(JetstreamJT270MHRadio):
    '''Newer VV898E 1.06+ firmware that features dual memory banks'''
    VENDOR = "Leixen"
    MODEL = "VV-898E"
    VARIANT = "Dual Bank"

    @classmethod
    def match_model(cls, filedata, filename):
        return False



# Copyright 2022 Jim Unroe <rock.unroe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import struct
import logging

from chirp import chirp_common, directory, memmap
from chirp import bitwise, errors, util
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueInteger, RadioSettingValueList, \
    RadioSettingValueBoolean, RadioSettings

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
#seekto 0x0010;
struct {
  char name[10];      // 10-character Alpha Tag
} names[128];
#seekto 0x0810;
struct {
    u8 unknown01:5,
       dualwatch:1,
       workmode:2;
    u8 scanmode:2,
       unknown02:3,
       autokeylock:1,
       unknown03:2;
    u8 unknown04:1,
       beep:1,
       unknown05:1,
       voice:1,
       unknown06:1,
       noaa:1,
       batterysave:2;
    u8 squelchlevel;
    u8 unknown07:5,
       timeouttimer:2,
       unknown08:1;
    u8 alarmtype:1,
       voxdelay:3,
       voxlevel:4;
    u8 unknown10:2,
       backlight:2,
       unknown11:4;
    u8 unknown12[3];
    u8 unknown13:3,
       rogerbeep:1,
       unknown14:1,
       sidekey:3;
} settings;
#seekto 0x1000;
struct {
    lbcd rxfreq[4];
    lbcd txfreq[4];
    lbcd rxtone[2];
    lbcd txtone[2];
    u8 unknown1:4,
       compander:1,
       unknown2:1,
       highpower:1,
       unknown3:1;
    u8 unknown4:3,
       wide:1,
       scan:1,
       unknown5:1,
       bcl:2;
    u8 unknown6[2];
} memory[128];

"""

CMD_ACK = b"\x06"

ALARMTYPE_LIST = ["Local & Remote", "Local"]
BACKLIGHT_LIST = ["5 seconds", "10 seconds", "30 seconds", "Always"]
BATTERYSAVE_LIST = ["Off", "1:4", "1:8"]
SCANMODE_LIST = ["Time Operated", "Carrier Operated", "Search"]
SIDEKEY_LIST = ["Off", "Scan", "VOX", "2nd PTT", "NOAA", "Monitor"]
TIMEOUTTIMER_LIST = ["60 seconds", "120 seconds", "180 seconds"]
VOXDELAY_LIST = ["0.3", "0.5", "1", "1.5", "2", "2.5"]
VOXLEVEL_LIST = ["Off", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
WORKMODE_LIST = ["Frequencies", "Channel Numbers", "Names"]

BCL = ["Off", "Carrier", "QT/DCS"]

GMRS_FREQS1 = [462562500, 462587500, 462612500, 462637500, 462662500,
               462687500, 462712500]
GMRS_FREQS2 = [467562500, 467587500, 467612500, 467637500, 467662500,
               467687500, 467712500]
GMRS_FREQS3 = [462550000, 462575000, 462600000, 462625000, 462650000,
               462675000, 462700000, 462725000]
GMRS_FREQS = GMRS_FREQS1 + GMRS_FREQS2 + GMRS_FREQS3 * 2


def _enter_programming_mode(radio):
    serial = radio.pipe

    try:
        serial.write(b"\x02")
        time.sleep(0.01)
        serial.write(radio._magic)
        ack = serial.read(1)
    except:
        raise errors.RadioError("Error communicating with radio")

    if not ack:
        raise errors.RadioError("No response from radio")
    elif ack != CMD_ACK:
        raise errors.RadioError("Radio refused to enter programming mode")

    try:
        serial.write(b"M" + b"\x02")
        ident = serial.read(8)
    except:
        raise errors.RadioError("Error communicating with radio")

    if not ident.startswith(radio._fingerprint):
        LOG.debug(util.hexprint(ident))
        raise errors.RadioError("Radio returned unknown identification string")

    try:
        serial.write(CMD_ACK)
        ack = serial.read(1)
    except:
        raise errors.RadioError("Error communicating with radio")

    if ack != CMD_ACK:
        raise errors.RadioError("Radio refused to enter programming mode")


def _read_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">cHb", b'R', block_addr, block_size)
    expectedresponse = b"W" + cmd[1:]
    LOG.debug("Reading block %04x..." % (block_addr))

    try:
        serial.write(cmd)
        response = serial.read(4 + block_size)
        if response[:4] != expectedresponse:
            raise Exception("Error reading block %04x." % (block_addr))

        block_data = response[4:]
    except:
        raise errors.RadioError("Failed to read block at %04x" % block_addr)

    return block_data


def _write_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">cHb", b'W', block_addr, block_size)
    data = radio.get_mmap()[block_addr:block_addr + block_size]

    LOG.debug("Writing Data:")
    LOG.debug(util.hexprint(cmd + data))

    try:
        serial.write(cmd + data)
        if serial.read(1) != CMD_ACK:
            raise Exception("No ACK")
    except:
        raise errors.RadioError("Failed to send block "
                                "to radio at %04x" % block_addr)


def do_download(radio):
    LOG.debug("download")
    _enter_programming_mode(radio)

    data = b""

    status = chirp_common.Status()
    status.msg = "Cloning from radio"

    status.cur = 0
    status.max = radio._memsize

    for addr in range(0, radio._memsize, radio.BLOCK_SIZE):
        status.cur = addr + radio.BLOCK_SIZE
        radio.status_fn(status)

        block = _read_block(radio, addr, radio.BLOCK_SIZE)
        data += block

        LOG.debug("Address: %04x" % addr)
        LOG.debug(util.hexprint(block))

    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    status = chirp_common.Status()
    status.msg = "Uploading to radio"

    _enter_programming_mode(radio)

    status.cur = 0
    status.max = radio._memsize

    for start_addr, end_addr, block_size in radio._ranges:
        for addr in range(start_addr, end_addr, block_size):
            status.cur = addr + block_size
            radio.status_fn(status)
            _write_block(radio, addr, block_size)


class RB17P_Base(chirp_common.CloneModeRadio):
    """Base class for Retevis RB17P"""
    VENDOR = "Retevis"
    MODEL = "RB17P Base"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False
    BLOCK_SIZE = 0x40

    VALID_BANDS = [(400000000, 470000000)]
    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=5.00),
                    chirp_common.PowerLevel("Low", watts=0.50)]

    _magic = b"PGA588"
    _fingerprint = b"\xFF" * 8
    _upper = 128
    _gmrs = True
    _ranges = [(0x0000, 0x1800, 0x40),
               ]
    _memsize = 0x1800
    _valid_chars = chirp_common.CHARSET_ALPHANUMERIC + \
        "`~!@#$%^&*()-=_+[]\\{}|;':\",./<>?"

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.valid_modes = ["NFM", "FM"]  # 12.5 kHz, 25 kHz.
        rf.valid_skips = ["", "S"]
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.valid_power_levels = self.POWER_LEVELS
        rf.can_odd_split = True
        rf.has_rx_dtcs = True
        rf.has_ctone = True
        rf.has_cross = True
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->",
            "->DTCS",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "DTCS->DTCS"]
        rf.has_tuning_step = False
        rf.has_bank = False
        rf.has_name = True
        rf.valid_name_length = 10
        rf.valid_characters = self._valid_chars
        rf.memory_bounds = (1, 128)
        rf.valid_bands = self.VALID_BANDS
        rf.valid_tuning_steps = chirp_common.TUNING_STEPS

        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def sync_in(self):
        self._mmap = do_download(self)
        self.process_mmap()

    def sync_out(self):
        do_upload(self)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1])

    def _decode_tone(self, val):
        val = int(val)
        if val == 16665:
            return '', None, None
        elif val >= 12000:
            return 'DTCS', val - 12000, 'R'
        elif val >= 8000:
            return 'DTCS', val - 8000, 'N'
        else:
            return 'Tone', val / 10.0, None

    def _encode_tone(self, memval, mode, value, pol):
        if mode == '':
            memval[0].set_raw(0xFF)
            memval[1].set_raw(0xFF)
        elif mode == 'Tone':
            memval.set_value(int(value * 10))
        elif mode == 'DTCS':
            flag = 0x80 if pol == 'N' else 0xC0
            memval.set_value(value)
            memval[1].set_bits(flag)
        else:
            raise Exception("Internal error: invalid mode `%s'" % mode)

    def get_memory(self, number):
        _mem = self._memobj.memory[number - 1]
        _nam = self._memobj.names[number - 1]

        mem = chirp_common.Memory()

        mem.number = number
        mem.freq = int(_mem.rxfreq) * 10

        # We'll consider any blank (i.e. 0 MHz frequency) to be empty
        if mem.freq == 0:
            mem.empty = True
            return mem

        if _mem.rxfreq.get_raw() == b"\xFF\xFF\xFF\xFF":
            mem.freq = 0
            mem.empty = True
            return mem

        if _mem.txfreq.get_raw() == b"\xFF\xFF\xFF\xFF":
            mem.duplex = "off"
            mem.offset = 0
        elif int(_mem.rxfreq) == int(_mem.txfreq):
            mem.duplex = ""
            mem.offset = 0
        else:
            mem.duplex = int(_mem.rxfreq) > int(_mem.txfreq) and "-" or "+"
            mem.offset = abs(int(_mem.rxfreq) - int(_mem.txfreq)) * 10

        for char in _nam.name:
            if str(char) == "\xFF":
                char = " "  # may have 0xFF mid-name
            mem.name += str(char)
        mem.name = mem.name.rstrip()

        mem.mode = _mem.wide and "FM" or "NFM"

        mem.skip = not _mem.scan and "S" or ""

        txtone = self._decode_tone(_mem.txtone)
        rxtone = self._decode_tone(_mem.rxtone)
        chirp_common.split_tone_decode(mem, txtone, rxtone)

        mem.power = self.POWER_LEVELS[1 - _mem.highpower]

        mem.extra = RadioSettingGroup("Extra", "extra")
        rs = RadioSetting("bcl", "Busy Channel Lockout",
                          RadioSettingValueList(BCL, BCL[_mem.bcl]))
        mem.extra.append(rs)
        rs = RadioSetting("compander", "Compander",
                          RadioSettingValueBoolean(_mem.compander))
        mem.extra.append(rs)

        immutable = []

        if self._gmrs:
            if mem.freq in GMRS_FREQS:
                if mem.freq in GMRS_FREQS1:
                    # Non-repeater GMRS channels (limit duplex)
                    mem.duplex == ''
                    mem.offset = 0
                    immutable = ["duplex", "offset"]
                elif mem.freq in GMRS_FREQS2:
                    # Non-repeater FRS channels (limit duplex, power)
                    mem.duplex == ''
                    mem.offset = 0
                    mem.mode = "NFM"
                    mem.power = self.POWER_LEVELS[1]
                    immutable = ["duplex", "offset", "mode", "power"]
                elif mem.freq in GMRS_FREQS3:
                    # GMRS repeater channels, always either simplex or +5 MHz
                    if mem.duplex != '+':
                        mem.duplex = ''
                        mem.offset = 0
                    else:
                        mem.offset = 5000000
            else:
                # Not a GMRS channel, so restrict duplex since it will be
                # forced to off.
                mem.duplex = 'off'
                mem.offset = 0
                immutable = ["duplex", "offset"]

        mem.immutable = immutable

        return mem

    def check_set_memory_immutable_policy(self, existing, new):
        existing.immutable = []
        super().check_set_memory_immutable_policy(existing, new)

    def set_memory(self, mem):
        # Get a low-level memory object mapped to the image
        _mem = self._memobj.memory[mem.number - 1]
        _nam = self._memobj.names[mem.number - 1]

        if mem.empty:
            _mem.set_raw(b"\xFF" * (_mem.size() // 8))
            _nam.set_raw(b"\xFF" * (_nam.size() // 8))

            return

        _mem.set_raw(b"\x00" * (_mem.size() // 8))

        _mem.rxfreq = mem.freq / 10

        if mem.duplex == "off":
            for i in range(0, 4):
                _mem.txfreq[i].set_raw(b"\xFF")
        elif mem.duplex == "split":
            _mem.txfreq = mem.offset / 10
        elif mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        else:
            _mem.txfreq = mem.freq / 10

        _namelength = self.get_features().valid_name_length
        for i in range(_namelength):
            try:
                _nam.name[i] = mem.name[i]
            except IndexError:
                _nam.name[i] = "\xFF"

        txtone, rxtone = chirp_common.split_tone_encode(mem)
        self._encode_tone(_mem.txtone, *txtone)
        self._encode_tone(_mem.rxtone, *rxtone)

        if self.MODEL != "T18" and self.MODEL != "RB18":
            _mem.highpower = mem.power == self.POWER_LEVELS[0]

        _mem.wide = mem.mode == "FM"
        _mem.scan = mem.skip == ""

        for setting in mem.extra:
            setattr(_mem, setting.get_name(), int(setting.value))

    def get_settings(self):
        _settings = self._memobj.settings
        basic = RadioSettingGroup("basic", "Basic Settings")
        top = RadioSettings(basic)

        # Menu 01 - Work Mode
        rs = RadioSetting("workmode", "Work Mode",
                          RadioSettingValueList(
                              WORKMODE_LIST,
                              WORKMODE_LIST[(_settings.workmode) - 1]))
        basic.append(rs)

        # Menu 04 - Squelch
        rs = RadioSetting("squelchlevel", "Squelch level",
                          RadioSettingValueInteger(
                              0, 9, _settings.squelchlevel))
        basic.append(rs)

        # Menu 05 - Battery Save
        rs = RadioSetting("batterysave", "Battery Save",
                          RadioSettingValueList(
                              BATTERYSAVE_LIST,
                              BATTERYSAVE_LIST[_settings.batterysave]))
        basic.append(rs)

        # Menu 06 - Dual Watch
        rs = RadioSetting("dualwatch", "Dual Watch",
                          RadioSettingValueBoolean(_settings.dualwatch))
        basic.append(rs)

        # Menu 07 - Backlight Duration[s]
        rs = RadioSetting("backlight", "Backlight Duration",
                          RadioSettingValueList(
                              BACKLIGHT_LIST,
                              BACKLIGHT_LIST[_settings.backlight]))
        basic.append(rs)

        # Menu 09 - Beep Tone
        rs = RadioSetting("beep", "Key Beep",
                          RadioSettingValueBoolean(_settings.beep))
        basic.append(rs)

        # Menu 10 - Voice Annunciation
        rs = RadioSetting("voice", "Voice prompts",
                          RadioSettingValueBoolean(_settings.voice))
        basic.append(rs)

        # Menu 11 - Time out Timer[s]
        rs = RadioSetting("timeouttimer", "Timeout timer",
                          RadioSettingValueList(
                              TIMEOUTTIMER_LIST,
                              TIMEOUTTIMER_LIST[
                                  (_settings.timeouttimer) - 1]))
        basic.append(rs)

        # Menu 12 - Roger Tone
        rs = RadioSetting("rogerbeep", "Roger beep",
                          RadioSettingValueBoolean(_settings.rogerbeep))
        basic.append(rs)

        # Menu 13 - SideKey
        rs = RadioSetting("sidekey", "Side Key",
                          RadioSettingValueList(
                              SIDEKEY_LIST,
                              SIDEKEY_LIST[_settings.sidekey]))
        basic.append(rs)

        # Menu 14 - Auto Key Lock
        rs = RadioSetting("autokeylock", "Auto Key Lock",
                          RadioSettingValueBoolean(_settings.autokeylock))
        basic.append(rs)

        # Menu 17 - NOAA
        rs = RadioSetting("noaa", "NOAA",
                          RadioSettingValueBoolean(_settings.noaa))
        basic.append(rs)

        # Scan Mode
        rs = RadioSetting("scanmode", "Scan mode",
                          RadioSettingValueList(
                              SCANMODE_LIST,
                              SCANMODE_LIST[_settings.scanmode]))
        basic.append(rs)

        # Alarm Type
        rs = RadioSetting("alarmtype", "Alarm Type",
                          RadioSettingValueList(
                              ALARMTYPE_LIST,
                              ALARMTYPE_LIST[_settings.alarmtype]))
        basic.append(rs)

        # VOX Level
        rs = RadioSetting("voxlevel", "Vox level",
                          RadioSettingValueList(
                              VOXLEVEL_LIST,
                              VOXLEVEL_LIST[_settings.voxlevel]))
        basic.append(rs)

        # VOX Delay
        rs = RadioSetting("voxdelay", "Vox delay",
                          RadioSettingValueList(
                              VOXDELAY_LIST,
                              VOXDELAY_LIST[_settings.voxdelay]))
        basic.append(rs)

        return top

    def set_settings(self, settings):
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    if "." in element.get_name():
                        bits = element.get_name().split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = self._memobj.settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif setting == "timeouttimer":
                        setattr(obj, setting, int(element.value) + 1)
                    elif setting == "workmode":
                        setattr(obj, setting, int(element.value) + 1)
                    else:
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    @classmethod
    def match_model(cls, filedata, filename):
        # Radios that have always been post-metadata, so never do
        # old-school detection
        return False


@directory.register
class RB17PRadio(RB17P_Base):
    """Retevis RB17P"""
    VENDOR = "Retevis"
    MODEL = "RB17P"
    _gmrs = True
    _ranges = [(0x0010, 0x0510, 0x40),
               (0x0810, 0x0830, 0x30),
               (0x1000, 0x1800, 0x40)
               ]

    def validate_memory(self, mem):
        msgs = super().validate_memory(mem)

        _msg_duplex = 'Duplex must be "off" for this frequency'
        _msg_offset = 'Only simplex or +5 MHz offset allowed on GMRS'

        if mem.freq not in GMRS_FREQS:
            if mem.duplex != "off":
                msgs.append(chirp_common.ValidationWarning(_msg_duplex))
        elif mem.duplex and mem.offset != 5000000:
            msgs.append(chirp_common.ValidationWarning(_msg_offset))

        return msgs



# Copyright 2012 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from chirp import chirp_common

LOG = logging.getLogger(__name__)
BANNED_NAME_CHARACTERS = r'%'


class InvalidValueError(Exception):

    """An invalid value was specified for a given setting"""
    pass


class InternalError(Exception):

    """A driver provided an invalid settings object structure"""
    pass


class InvalidNameError(Exception):
    """A driver provided a setting with an invalid name"""
    pass


class RadioSettingValue:

    """Base class for a single radio setting"""

    def __init__(self):
        self._current = None
        self._preinit_current = None
        self._has_changed = False
        self._validate_callback = self.null_callback
        self._mutable = True

    def initialize(self):
        """Initialize this value to the stashed value"""
        assert self._current is None and self._preinit_current is not None
        self.set_value(self._preinit_current)
        self._preinit_current = None

    def queue_current(self, current):
        """Set this value's current on initialization"""
        self._preinit_current = current

    def null_callback(self, x):
        return x

    def set_mutable(self, mutable):
        self._mutable = mutable

    def get_mutable(self):
        return self._mutable

    def changed(self):
        """Returns True if the setting has been changed since init"""
        return self._has_changed

    def set_validate_callback(self, callback):
        self._validate_callback = callback

    def set_value(self, value):
        """Sets the current value, triggers changed"""
        if not self.get_mutable() and self.initialized:
            raise InvalidValueError("This value is not mutable")

        if self._current is not None and value != self._current:
            self._has_changed = True
        self._current = self._validate_callback(value)

    def get_value(self):
        """Gets the current value"""
        if not self.initialized:
            LOG.error('Accessing setting value before init!')
        return self._current

    def __trunc__(self):
        return int(self.get_value())

    def __int__(self):
        return int(self.get_value())

    def __float__(self):
        return float(self.get_value())

    def __str__(self):
        return str(self.get_value())

    def __floordiv__(self, value):
        return int(self.get_value()) // value

    def __div__(self, value):
        return float(self.get_value()) / value

    def __mul__(self, value):
        return self.get_value() * value

    def __gt__(self, value):
        return self.get_value() > value

    def __lt__(self, value):
        return self.get_value() < value

    def __eq__(self, value):
        return self.get_value() == value

    @property
    def initialized(self):
        return self._current is not None


class RadioSettingValueInteger(RadioSettingValue):

    """An integer setting"""

    def __init__(self, minval, maxval, current, step=1):
        RadioSettingValue.__init__(self)
        self._min = minval
        self._max = maxval
        self._step = step
        self.queue_current(current)

    def set_value(self, value):
        try:
            value = int(value)
        except Exception:
            raise InvalidValueError("An integer is required")
        if value > self._max or value < self._min:
            raise InvalidValueError("Value %i not in range %i-%i" %
                                    (value, self._min, self._max))
        RadioSettingValue.set_value(self, value)

    def get_min(self):
        """Returns the minimum allowed value"""
        return self._min

    def get_max(self):
        """Returns the maximum allowed value"""
        return self._max

    def get_step(self):
        """Returns the step increment"""
        return self._step


class RadioSettingValueFloat(RadioSettingValue):

    """A floating-point setting"""

    def __init__(self, minval, maxval, current, resolution=0.001, precision=4):
        RadioSettingValue.__init__(self)
        self._min = minval
        self._max = maxval
        self._res = resolution
        self._pre = precision
        self.queue_current(current)

    def format(self, value=None):
        """Formats the value into a string"""
        if value is None:
            value = self._current
        fmt_string = "%%.%if" % self._pre
        return fmt_string % value

    def set_value(self, value):
        try:
            value = float(value)
        except Exception:
            raise InvalidValueError("A floating point value is required")
        if value > self._max or value < self._min:
            raise InvalidValueError("Value %s not in range %s-%s" % (
                self.format(value),
                self.format(self._min), self.format(self._max)))

        # FIXME: honor resolution

        RadioSettingValue.set_value(self, value)

    def get_min(self):
        """Returns the minimum allowed value"""
        return self._min

    def get_max(self):
        """Returns the maximum allowed value"""


class RadioSettingValueBoolean(RadioSettingValue):

    """A boolean setting"""

    def __init__(self, current):
        RadioSettingValue.__init__(self)
        self.queue_current(current)

    def set_value(self, value):
        RadioSettingValue.set_value(self, bool(value))

    def __bool__(self):
        return bool(self.get_value())
    __nonzero__ = __bool__

    def __str__(self):
        return str(bool(self.get_value()))


class RadioSettingValueList(RadioSettingValue):

    """A list-of-strings setting"""

    def __init__(self, options, current=None, current_index=0):
        RadioSettingValue.__init__(self)
        self._options = list(options)
        self.queue_current(current or self._options[int(current_index)])

    def set_value(self, value):
        if value not in self._options:
            raise InvalidValueError("%s is not valid for this setting" % value)
        RadioSettingValue.set_value(self, value)

    def set_index(self, index):
        # Will raise IndexError as expected
        self.set_value(self._options[int(index)])

    def get_options(self):
        """Returns the list of valid option values"""
        return self._options

    def __trunc__(self):
        return self._options.index(self._current)

    def __int__(self):
        return self._options.index(self._current)


class RadioSettingValueString(RadioSettingValue):

    """A string setting"""

    def __init__(self, minlength, maxlength, current,
                 autopad=True, charset=chirp_common.CHARSET_ASCII):
        RadioSettingValue.__init__(self)
        self._minlength = minlength
        self._maxlength = maxlength
        self._charset = charset
        self._autopad = autopad
        self.queue_current(current)

    def set_charset(self, charset):
        """Sets the set of allowed characters"""
        self._charset = charset

    def set_value(self, value):
        if len(value) < self._minlength or len(value) > self._maxlength:
            raise InvalidValueError("Value must be between %i and %i chars" %
                                    (self._minlength, self._maxlength))
        for char in value:
            if char not in self._charset:
                raise InvalidValueError("Value contains invalid " +
                                        "character `%s'" % char)
        if self._autopad:
            value = value.ljust(self._maxlength)
        RadioSettingValue.set_value(self, value)

    def __str__(self):
        return self._current

    def __len__(self):
        return len(self._current)

    def __getitem__(self, i):
        return self._current[i]


class RadioSettingValueMap(RadioSettingValueList):

    """Map User Options to Radio Memory Values

    Provides User Option list for GUI, maintains state, verifies new values,
    and allows {setting,getting} by User Option OR Memory Value.  External
    conversions not needed.

    """

    def __init__(self, map_entries, mem_val=None, user_option=None):
        """Create new map

        Pass in list of 2 member tuples, typically of type (str, int),
        for each Radio Setting.  First member of each tuple is the
        User Option Name, second is the Memory Value that corresponds.
        An example is APO: ("Off", 0), ("0.5", 5), ("1.0", 10).

        """
        # Catch bugs early by testing tuple geometry
        for map_entry in map_entries:
            if not len(map_entry) == 2:
                raise InvalidValueError("map_entries must be 2 el tuples "
                                        "instead of: %s" % str(map_entry))
        user_options = [e[0] for e in map_entries]
        self._mem_vals = [e[1] for e in map_entries]
        RadioSettingValueList.__init__(self, user_options, user_options[0])
        if mem_val is not None:
            self.set_mem_val(mem_val)
        elif user_option is not None:
            self.set_value(user_option)
        self._has_changed = False

    def set_mem_val(self, mem_val):
        """Change setting to User Option that corresponds to 'mem_val'"""
        if mem_val in self._mem_vals:
            index = self._mem_vals.index(mem_val)
            self.set_value(self._options[index])
        else:
            raise InvalidValueError(
                "%s is not valid for this setting" % mem_val)

    def get_mem_val(self):
        """Get the mem val corresponding to the currently selected user
        option"""
        return self._mem_vals[self._options.index(self.get_value())]

    def __trunc__(self):
        """Return memory value that matches current user option"""
        index = self._options.index(self._current)
        value = self._mem_vals[index]
        return value

    def __int__(self):
        """Return memory value that matches current user option"""
        index = self._options.index(self._current)
        value = self._mem_vals[index]
        return value


def zero_indexed_seq_map(user_options):
    """RadioSettingValueMap factory method

    Radio Setting Maps commonly use a list of strings that map to a sequence
    that starts with 0.  Pass in a list of User Options and this function
    returns a list of tuples of form (str, int).

    """
    mem_vals = list(range(0, len(user_options)))
    return list(zip(user_options, mem_vals))


class RadioSettings(list):

    def __init__(self, *groups):
        list.__init__(self, groups)

    def __str__(self):
        items = [str(self[i]) for i in range(0, len(self))]
        return "\n".join(items)


class RadioSettingGroup(object):

    """A group of settings"""

    def _validate(self, element):
        # RadioSettingGroup can only contain RadioSettingGroup objects
        if not isinstance(element, RadioSettingGroup):
            raise InternalError("Incorrect type %s" % type(element))

    def __init__(self, name, shortname, *elements):
        for c in BANNED_NAME_CHARACTERS:
            if c in name:
                raise InvalidNameError(
                    'Name must not contain %r character' % c)
        self._name = name            # Setting identifier
        self._shortname = shortname  # Short human-readable name/description
        self.__doc__ = name          # Longer explanation/documentation
        self._elements = {}
        self._element_order = []
        self._frozen = False

        for i, element in enumerate(elements):
            self._validate(element)
            if (isinstance(element, RadioSettingValue) and
                    not element.initialized):
                # Late-initialize this value
                try:
                    element.initialize()
                except InvalidValueError as e:
                    LOG.error('Failed to load %s value %i: %s',
                              self._name, i, e)

            self.append(element)

    def set_frozen(self):
        self._frozen = True
        for i in self:
            if isinstance(i, RadioSettingGroup):
                i.set_frozen()

    def get_name(self):
        """Returns the group name"""
        return self._name

    def get_shortname(self):
        """Returns the short group identifier"""
        return self._shortname

    def set_doc(self, doc):
        """Sets the docstring for the group"""
        self.__doc__ = doc

    def __str__(self):
        string = "group '%s': {\n" % self._name
        for element in sorted(self._elements.values()):
            for line in str(element).split("\n"):
                string += "\t" + line + "\n"
        string += "}"
        return string

    # Kinda list interface

    def append(self, element):
        """Adds an element to the group"""
        if self._frozen:
            raise ValueError('Setting is frozen')
        self[element.get_name()] = element

    def __iter__(self):
        class RSGIterator:

            """Iterator for a RadioSettingGroup"""

            def __init__(self, rsg):
                self.__rsg = rsg
                self.__i = 0

            def __iter__(self):
                return self

            def next(self):
                return self.__next__()

            def __next__(self):
                """Next Iterator Interface"""
                if self.__i >= len(self.__rsg.keys()):
                    raise StopIteration()
                e = self.__rsg[self.__rsg.keys()[self.__i]]
                self.__i += 1
                return e
        return RSGIterator(self)

    # Dictionary interface

    def __len__(self):
        return len(self._elements)

    def __getitem__(self, name):
        return self._elements[name]

    def __setitem__(self, name, value):
        if name in self._element_order:
            raise KeyError("Duplicate item %s" % name)
        self._elements[name] = value
        self._element_order.append(name)

    def __delitem__(self, item):
        del self._elements[item.get_name()]
        self._element_order.remove(item.get_name())

    def __contains__(self, name):
        return name in self._elements

    def items(self):
        """Returns a key=>value set of elements, like a dict"""
        return [(name, self._elements[name]) for name in self._element_order]

    def keys(self):
        """Returns a list of string element names"""
        return self._element_order

    def values(self):
        """Returns the list of elements"""
        return [self._elements[name] for name in self._element_order]

    def __lt__(self, other):
        return self._name < other._name


class RadioSettingSubGroup(RadioSettingGroup):
    """A subgroup of related settings.

    These are to be displayed with less weight than a regular group. Used
    for repeated sections of related settings, or small groups that don't
    merit their own panel of display in the UI.
    """


class RadioSetting(RadioSettingGroup):

    """A single setting, which could be an array of items like a group"""

    def __init__(self, *args):
        super(RadioSetting, self).__init__(*args)
        self._apply_callback = None
        self._warning_text = None
        self._safe_value = None
        self._volatile = False

    @property
    def volatile(self):
        return self._volatile

    def set_volatile(self, value):
        self._volatile = value

    def set_apply_callback(self, callback, *args):
        self._apply_callback = lambda: callback(self, *args)

    def has_apply_callback(self):
        return self._apply_callback is not None

    def run_apply_callback(self):
        return self._apply_callback()

    def set_warning(self, warning_text, safe_value=None):
        """Set a warning message to be shown to the user when changed.

        This should not be over-used as it will be annoying (on purpose). This
        message will be shown to the user with an option to continue to abort.
        It should be used to warn the user of potential invalid operation or
        compatibility issues. If safe_value is set, this will only warn the
        user if the value is changed *to* something other than the safe_value.
        """
        self._warning_text = warning_text
        self._safe_value = safe_value

    def get_warning(self, value):
        if value != self._safe_value:
            return self._warning_text

    def _validate(self, value):
        # RadioSetting can only contain RadioSettingValue objects
        if not isinstance(value, RadioSettingValue):
            raise InternalError("Incorrect type")

    def changed(self):
        """Returns True if any of the elements
        in the group have been changed"""
        for element in self._elements.values():
            if element.changed():
                return True
        return False

    def __str__(self):
        return "%s:%s" % (self._name, self.value)

    def __repr__(self):
        return "[RadioSetting %s:%s]" % (self._name, self.value)

    # Magic foo.value attribute
    def __getattr__(self, name):
        if name == "value":
            if len(self) == 1:
                return self._elements[self._element_order[0]]
            else:
                return list(self._elements.values())
        elif name in ('__getstate__', '__setstate__', '__deepcopy__'):
            super().__getattr__(name)
        else:
            return self.__dict__[name]

    def __setattr__(self, name, value):
        if name == "value":
            if self._frozen:
                raise ValueError('Value is frozen')
            if len(self) == 1:
                self._elements[self._element_order[0]].set_value(value)
            else:
                raise InternalError("Setting %s is not a scalar" % self._name)
        else:
            self.__dict__[name] = value

    # List interface

    def append(self, value):
        index = len(self._element_order)
        self._elements[index] = value
        self._element_order.append(index)

    def __getitem__(self, name):
        if not isinstance(name, int):
            raise IndexError("Index `%s' is not an integer" % name)
        return self._elements[name]

    def __setitem__(self, name, value):
        if not isinstance(name, int):
            raise IndexError("Index `%s' is not an integer" % name)
        if name in self._elements:
            self._elements[name].set_value(value)
        else:
            self._elements[name] = value



# Copyright 2023 Dan Smith <chirp@f.danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

import requests

from chirp.drivers import generic_csv
from chirp import errors
from chirp.sources import base

LOG = logging.getLogger(__name__)


class Przemienniki(base.NetworkResultRadio):
    VENDOR = 'przemienniki.net'

    def get_label(self):
        return 'przemienniki.net'

    def do_fetch(self, status, params):
        status.send_status('Querying', 10)
        if not params['range'] or int(params['range']) == 0:
            params.pop('range')
        LOG.debug('query params: %s' % str(params))
        try:
            r = requests.get('http://przemienniki.net/export/chirp.csv',
                             headers=base.HEADERS,
                             params=params,
                             stream=True)
            r.raise_for_status()
        except requests.exceptions.RequestException as e:
            LOG.error('Failed to query przemienniki: %s' % e)
            status.send_fail(_('Unable to query'))
            return
        status.send_status(_('Parsing'), 20)
        try:
            csv = generic_csv.CSVRadio(None)
            csv._load(x.decode() for x in r.iter_lines())
        except errors.InvalidDataError:
            status.send_fail(_('No results'))
            return
        except Exception as e:
            LOG.error('Error parsing result: %s' % e)
            status.send_fail(_('Failed to parse result'))
            return

        status.send_status(_('Sorting'), 80)

        self._memories = [csv.get_memory(x) for x in range(1, 999)
                          if not csv.get_memory(x).empty]
        if not any([params['latitude'], params['longitude']]):
            LOG.debug('Sorting memories by name')
            self._memories.sort(key=lambda m: m.name)
            # Now renumber them
            for i, mem in enumerate(self._memories):
                mem.number = i + 1

        return status.send_end()



# Copyright 2008 Dan Smith <dsmith@danplanet.com>
# Copyright 2012 Tom Hayward <tom@tomh.us>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import csv
import logging

from chirp import chirp_common, errors, directory
from chirp.drivers import generic_csv

LOG = logging.getLogger(__name__)


class OmittedHeaderError(Exception):
    """An internal exception to indicate that a header was omitted"""
    pass


@directory.register
class HMKRadio(generic_csv.CSVRadio):
    """Kenwood HMK format"""
    VENDOR = "Kenwood"
    MODEL = "HMK"
    FILE_EXTENSION = "hmk"

    DUPLEX_MAP = {
        " ":    "",
        "S":    "split",
        "+":    "+",
        "-":    "-",
    }

    SKIP_MAP = {
        "Off":  "",
        "On":   "S",
        }

    TMODE_MAP = {
        "Off":  "",
        "T":    "Tone",
        "CT":   "TSQL",
        "DCS":  "DTCS",
        "":     "Cross",
        }

    ATTR_MAP = {
        "!!Ch":          (int,   "number"),
        "M.Name":        (str,   "name"),
        "Rx Freq.":      (chirp_common.parse_freq, "freq"),
        "Shift/Split":   (lambda v: HMKRadio.DUPLEX_MAP[v], "duplex"),
        "Offset":        (chirp_common.parse_freq, "offset"),
        "T/CT/DCS":      (lambda v: HMKRadio.TMODE_MAP[v], "tmode"),
        "TO Freq.":      (float, "rtone"),
        "CT Freq.":      (float, "ctone"),
        "DCS Code":      (int,   "dtcs"),
        "Mode":          (str,   "mode"),
        "Rx Step":       (float, "tuning_step"),
        "L.Out":         (lambda v: HMKRadio.SKIP_MAP[v], "skip"),
        }

    def load(self, filename=None):
        if filename is None and self._filename is None:
            raise errors.RadioError("Need a location to load from")

        if filename:
            self._filename = filename

        self._blank()

        f = open(self._filename, "r")
        for line in f:
            if line.strip() == "// Memory Channels":
                break

        reader = csv.reader(f, delimiter=chirp_common.SEPCHAR, quotechar='"')

        good = 0
        lineno = 0
        for line in reader:
            lineno += 1
            if lineno == 1:
                header = line
                continue

            if len(header) > len(line):
                LOG.debug("Line %i has %i columns, expected %i" %
                          (lineno, len(line), len(header)))
                self.errors.append("Column number mismatch on line %i" %
                                   lineno)
                continue

            # hmk stores Tx Freq. in its own field, but Chirp expects the Tx
            # Freq. for odd-split channels to be in the Offset field.
            # If channel is odd-split, copy Tx Freq. field to Offset field.
            if line[header.index('Shift/Split')] == "S":
                line[header.index('Offset')] = line[header.index('Tx Freq.')]

            # fix EU decimal
            line = [i.replace(',', '.') for i in line]

            try:
                mem = self._parse_csv_data_line(header, line)
                if mem.number is None:
                    raise Exception("Invalid Location field" % lineno)
            except Exception as e:
                LOG.error("Line %i: %s" % (lineno, e))
                self.errors.append("Line %i: %s" % (lineno, e))
                continue

            self._grow(mem.number)
            self.memories[mem.number] = mem
            good += 1

        if not good:
            for error in errors:
                LOG.error("kenwood_hmk: %s", error)
            raise errors.InvalidDataError("No channels found")

    @classmethod
    def match_model(cls, filedata, filename):
        """Match files ending in .hmk"""
        return filename.lower().endswith("." + cls.FILE_EXTENSION)



# Copyright 2016 Tom Hayward <tom@tomh.us>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import requests
from chirp import chirp_common
from chirp.sources import base
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueList

LOG = logging.getLogger(__name__)


def list_filter(haystack, attr, needles):
    if not needles or not needles[0]:
        return haystack
    return [x for x in haystack if x[attr] in needles]


class DMRMARCRadio(base.NetworkResultRadio):
    """DMR-MARC data source"""
    VENDOR = "DMR-MARC"

    def get_label(self):
        return 'DMR-MARC'

    def do_fetch(self, status, params):
        status.send_status('Querying', 10)
        try:
            r = requests.get('https://radioid.net/api/dmr/repeater/',
                             headers=base.HEADERS,
                             params={'city': params['city'],
                                     'state': params['state'],
                                     'country': params['country']})
            r.raise_for_status()
        except requests.exceptions.RequestException as e:
            LOG.error('Failed to query DMR-MARC: %s' % e)
            status.send_fail('Unable to query DMR-MARC')
            return
        status.send_status('Parsing', 20)
        self._repeaters = r.json()['results']
        self._memories = [self.make_memory(i)
                          for i in range(0, len(self._repeaters))]
        status.send_end()

    def get_raw_memory(self, number):
        return repr(self._repeaters[number])

    def make_memory(self, number):
        repeater = self._repeaters[number]

        mem = chirp_common.Memory()
        mem.number = number

        mem.name = repeater.get('city')
        mem.freq = chirp_common.parse_freq(repeater.get('frequency'))
        offset = chirp_common.parse_freq(repeater.get('offset', '0'))
        if offset > 0:
            mem.duplex = "+"
        elif offset < 0:
            mem.duplex = "-"
        else:
            mem.duplex = ""
        mem.offset = abs(offset)
        mem.mode = 'DMR'
        mem.comment = repeater.get('details')

        mem.extra = RadioSettingGroup("Extra", "extra")

        rs = RadioSetting(
            "color_code", "Color Code", RadioSettingValueList(
                range(16), int(repeater.get('color_code', 0))))
        mem.extra.append(rs)

        return mem


def main():
    import argparse
    from pprint import PrettyPrinter

    parser = argparse.ArgumentParser(
        description=("Fetch DMR-MARC repeater "
                     "database and filter by city, state, and/or country. "
                     "Multiple items combined with a , will be filtered with "
                     "logical OR."))
    parser.add_argument(
        "-c", "--city",
        help="Comma-separated list of cities to include in output.")
    parser.add_argument(
        "-s", "--state",
        help="Comma-separated list of states to include in output.")
    parser.add_argument(
        "--country",
        help="Comma-separated list of countries to include in output.")
    args = parser.parse_args()

    dmrmarc = DMRMARCRadio(None)
    dmrmarc.set_params(**vars(args))
    dmrmarc.do_fetch()
    pp = PrettyPrinter(indent=2)
    pp.pprint(dmrmarc._repeaters)


if __name__ == "__main__":
    main()
